; Z88 CamelForth
; camel.s (mostly core primitives and inner interpreter)

include "camelusr.def"

; INTERPRETER LOGIC =============================
; See also "defining words" at end of this file

;C EXIT     --      exit a colon definition
    head EXIT,4,"EXIT",docode,0,0
        ld e,(ix+0)    ; pop old IP from ret stk
        inc ix
        ld d,(ix+0)
        inc ix
        innext		; inline NEXT for speed


;Z NOOP		--	do nothing
   head NOOP,4,"NOOP",docode,1,0
	next


;Z LIT      -- x    fetch inline literal to stack
; This is the primtive compiled by LITERAL.
    nohead LIT,docode
        push bc        ; push old TOS
        ld a,(de)      ; fetch cell at IP to TOS,
        ld c,a         ;        advancing IP
        inc de
        ld a,(de)
        ld b,a
        inc de
        innext		; inline NEXT for speed

;C EXECUTE   i*x xt -- j*x   execute Forth word
;C                           at 'xt'
    head EXECUTE,7,"EXECUTE",docode,0,0
        ld h,b          ; address of word -> HL
        ld l,c
        pop bc          ; get new TOS
        jp (hl)         ; go do Forth word

; DEFINING WORDS ================================

; ENTER, a.k.a. DOCOLON, entered by CALL ENTER
; to enter a new high-level thread (colon def'n.)
; (internal code fragment, not a Forth word)
; N.B.: DOCOLON must be defined before any
; appearance of 'docolon' in a 'word' macro!
.docolon               ; (alternate name)
.enter  dec ix         ; push old IP on ret stack
        ld (ix+0),d
        dec ix
        ld (ix+0),e
	pop de		; IP=body of colon definition
			; continue on into NEXT

; The heart of FORTH! The NEXT engine

.next_code
        ex de,hl
.nexthl_code
        ld e,(hl)
        inc hl
        ld d,(hl)
        inc hl
        ex de,hl
        jp (hl)


; The action of all CREATEd words

.docreate  ; -- x
        pop hl     ; parameter field address
        push bc    ; push old TOS
        ld c,(hl)  ; fetch contents of parameter
        inc hl     ;    field -> TOS
        ld b,(hl)
        innext		; inline NEXT for speed

;Z DEFER	"name" --	define deferred word
;   CREATE ['] NOOP , DOES> (code fragment)
    head DEFER,5,"DEFER",docolon,0,0
	defw CREATE,LIT,NOOP,COMMA,XDOES
.dodefer        ; i*x -- j*x
	pop hl		; pointer to data field
	push de		; save IP
	ld e,(hl)
	inc hl
	ld h,(hl)
	ld l,e		; HL=address of xt
	ld e,(hl)
	inc hl
	ld d,(hl)	; DE=xt
	ex de,hl
	pop de		; restore IP
	jp (hl)		; execute word


;C CONSTANT 	x --	define constant
;   NSDS CREATE -1 CELLS ALLOT , DOES> (code fragment)
; Constants don't need a pointer in their parameter field, as
; they should always be compiled with DATASPACE=NAMESPACE
    head CONSTANT,8,"CONSTANT",docolon,0,0
	defw NSDS,CREATE,LIT,-2,ALLOT,COMMA,XDOES
.docon                  ; -- x
	pop hl		; pointer to data field
	push bc		; push old TOS
	ld c,(hl)
	inc hl
	ld b,(hl)	; TOS=constant
	innext		; inline NEXT for speed

; DODOES, code action of DOES> clause
; entered by       CALL fragment
;                  defw  parameter field
;                       ...
;        fragment: CALL DODOES
;                  high-level thread
; Enters high-level thread with address of
; address parameter field on top of stack.
; (internal code fragment, not a Forth word)
.dodoes ; -- a-addr
        dec ix         ; push old IP on ret stk
        ld (ix+0),d
        dec ix
        ld (ix+0),e
        pop de         ; adrs of new thread -> IP
        pop hl         ; pointer to parameter field
        push bc        ; push old TOS onto stack
        ld c,(hl)      ; pfa -> new TOS
	inc hl
        ld b,(hl)
        innext		; inline NEXT for speed



; STACK OPERATIONS ==============================

;C DUP      x -- x x      duplicate top of stack
    head DUP,3,"DUP",docode,0,dupn-dupf
.dupf   push bc
.dupn
        next

;C ?DUP     x -- 0 | x x    DUP if nonzero
    head QDUP,4,"?DUP",docode,0,qdupn-qdupf
.qdupf  ld a,b
        or c
        jr z,qdupn
	push bc
.qdupn
        next

;C DROP     x --          drop top of stack
    head DROP,4,"DROP",docode,0,dropn-dropf
.dropf  pop bc
.dropn
        next

;C SWAP     x1 x2 -- x2 x1    swap top two items
    head SWOP,4,"SWAP",docode,0,swopn-swopf
.swopf  pop hl
        push bc
        ld b,h
        ld c,l
.swopn
        next

;C OVER    x1 x2 -- x1 x2 x1   per stack diagram
    head OVER,4,"OVER",docode,0,overn-overf
.overf  pop hl
        push hl
        push bc
        ld b,h
        ld c,l
.overn
        next

;C ROT    x1 x2 x3 -- x2 x3 x1  per stack diagram
    head ROT,3,"ROT",docode,0,rotn-rotf
        ; x3 is in TOS
.rotf   pop hl          ; x2
        ex (sp),hl      ; x2 on stack, x1 in hl
        push bc
        ld b,h
        ld c,l
.rotn
        next

;X NIP    x1 x2 -- x2           per stack diagram
    head NIP,3,"NIP",docode,0,nipn-nipf
.nipf   pop hl          ; discard second item
.nipn
        next

;X TUCK   x1 x2 -- x2 x1 x2     per stack diagram
    head TUCK,4,"TUCK",docode,0,tuckn-tuckf
.tuckf  pop hl
	push bc		; insert top item
	push hl
.tuckn
        next


;C >R    x --   R: -- x   push to return stack
    head TOR,2,">R",docode,0,torn-torf
.torf   dec ix          ; push TOS onto rtn stk
        ld (ix+0),b
        dec ix
        ld (ix+0),c
.roll2  pop bc          ; pop new TOS
.torn
        next

;X 2>R   x y --  R: -- x y  double push to return stack
    head TWOTOR,3,"2>R",docode,0,twotrn-twotrf
.twotrf pop hl
	dec ix
	ld (ix+0),h
	dec ix
	ld (ix+0),l
	dec ix
	ld (ix+0),b
	dec ix
	ld (ix+0),c
	pop bc		; pop new TOS
.twotrn
        next

;C R>    -- x    R: x --   pop from return stack
    head RFROM,2,"R>",docode,0,rfromn-rfromf
.rfromf push bc         ; push old TOS
        ld c,(ix+0)     ; pop top rtn stk item
        inc ix          ;       to TOS
        ld b,(ix+0)
        inc ix
.rfromn
        next

;X 2R>   -- x y  R: x y --   double pop from return stack
    head TWORFROM,3,"2R>",docode,0,tworfn-tworff
.tworff push bc         ; push old TOS
	ld c,(ix+0)
	inc ix
	ld b,(ix+0)	; top return item to BC=TOS
	inc ix
	ld l,(ix+0)
	inc ix
	ld h,(ix+0)
	inc ix
	push hl		; 2nd return item to 2OS
.tworfn
        next

;C R@    -- x     R: x -- x   fetch from rtn stk
    head RFETCH,2,"R@",docode,0,rfetcn-rfetcf
.rfetcf push bc         ; push old TOS
        ld c,(ix+0)     ; fetch top rtn stk item
        ld b,(ix+1)     ;       to TOS
.rfetcn
        next

;X 2R@   -- x y   R: x y -- x y  double fetch from return stack
    head TWORFETCH,3,"2R@",docode,0,twrftn-twrftf
.twrftf push bc         ; push old TOS
	ld c,(ix+2)
	ld b,(ix+3)
	push bc
	ld c,(ix+0)
	ld b,(ix+1)
.twrftn
        next

;Z SP@  -- a-addr       get data stack pointer
    nohead SPFETCH,docode
	push bc
        ld hl,0
        add hl,sp
        ld b,h
        ld c,l
        next

;Z SP!  a-addr --       set data stack pointer
    nohead SPSTORE,docode
	ld h,b
        ld l,c
        ld sp,hl
        pop bc          ; get new TOS
        next

;Z RP@  -- a-addr       get return stack pointer
    nohead RPFETCH,docode
	push bc
        push ix
        pop bc
        next

;Z RP!  a-addr --       set return stack pointer
    nohead RPSTORE,docode
	push bc
        pop ix
        pop bc
        next

; MEMORY AND I/O OPERATIONS =====================

;C !        x a-addr --   store cell in memory
    head STORE,1,"!",docode,0,storen-storef
.STOREDEST
.storef ld h,b          ; address in hl
        ld l,c
        pop bc          ; data in bc
        ld (hl),c
        inc hl
        ld (hl),b
        pop bc          ; pop new TOS
.storen
        next

;C C!      char c-addr --    store char in memory
    head CSTORE,2,"C!",docode,0,cstorn-cstorf
.cstorf ld h,b          ; address in hl
        ld l,c
        pop bc          ; data in bc
        ld (hl),c
        pop bc          ; pop new TOS
.cstorn
        next

;C @       a-addr -- x   fetch cell from memory
    head FETCH,1,"@",docode,0,fetchn-fetchf
.fetchf ld h,b          ; address in hl
        ld l,c
        ld c,(hl)
        inc hl
        ld b,(hl)
.fetchn
        next

;C C@     c-addr -- char   fetch char from memory
    head CFETCH,2,"C@",docode,0,cfetcn-cfetcf
.cfetcf ld a,(bc)
        ld c,a
        ld b,0
.cfetcn
        next

;Z PC!     char c-addr --    output char to port
    head PCSTORE,3,"PC!",docode,0,pcston-pcstof
.pcstof pop hl          ; char in L
        out (c),l       ; to port (BC)
        pop bc          ; pop new TOS
.pcston
        next

;Z PC@     c-addr -- char   input char from port
    head PCFETCH,3,"PC@",docode,0,pcfetn-pcfetf
.pcfetf in c,(c)        ; read port (BC) to C
        ld b,0
.pcfetn
        next

; ARITHMETIC AND LOGICAL OPERATIONS =============

;C +       n1/u1 n2/u2 -- n3/u3     add n1+n2
    head PLUS,1,"+",docode,0,plusn-plusf
.plusf  pop hl
        add hl,bc
        ld b,h
        ld c,l
.plusn
        next

;X M+       d n -- d         add single to double
    head MPLUS,2,"M+",docode,0,mplusn-mplusf
.mplusf ex de,hl
        pop de          ; hi cell
        ex (sp),hl      ; lo cell, save IP
        add hl,bc
        ld b,d          ; hi result in BC (TOS)
        ld c,e
        jr nc,mplus1
        inc bc
.mplus1 pop de          ; restore saved IP
        push hl         ; push lo result
.mplusn
        next

;X D+   d1 d2 -- d1+d2          add double numbers
    head DPLUS,2,"D+",docode,0,dplusn-dplusf
.dplusf exx
        pop bc          ; BC'=d2lo
        exx
        pop hl          ; HL=d1hi,BC=d2hi
        exx
        pop hl          ; HL'=d1lo
        add hl,bc
        push hl         ; 2OS=d1lo+d2lo
        exx
        adc hl,bc       ; HL=d1hi+d2hi+cy
        ld b,h
        ld c,l
.dplusn
        next

;X D-   d1 d2 -- d1-d2          subtract double numbers
    head DMINUS,2,"D-",docode,0,dminun-dminuf
.dminuf exx
        pop bc          ; BC'=d2lo
        exx
        pop hl          ; HL=d1hi,BC=d2hi
        exx
        pop hl          ; HL'=d1lo
        and a
        sbc hl,bc
        push hl         ; 2OS=d1lo-d2lo
        exx
        sbc hl,bc       ; HL=d1hi-d2hi-cy
        ld b,h
        ld c,l
.dminun
        next

;C -      n1/u1 n2/u2 -- n3/u3    subtract n1-n2
    head MINUS,1,"-",docode,0,minusn-minusf
.minusf pop hl
        or a
        sbc hl,bc
        ld b,h
        ld c,l
.minusn
        next

;C AND    x1 x2 -- x3            logical AND
    head AND,3,"AND",docode,0,andn-andf
.andf   pop hl
        ld a,b
        and h
        ld b,a
        ld a,c
        and l
        ld c,a
.andn
        next

;C OR     x1 x2 -- x3           logical OR
    head OR,2,"OR",docode,0,orn-orf
.orf    pop hl
        ld a,b
        or h
        ld b,a
        ld a,c
        or l
        ld c,a
.orn
        next

;C XOR    x1 x2 -- x3            logical XOR
    head XOR,3,"XOR",docode,0,xorn-xorf
.xorf   pop hl
        ld a,b
        xor h
        ld b,a
        ld a,c
        xor l
        ld c,a
.xorn
        next

;C INVERT   x1 -- x2            bitwise inversion
    head INVERT,6,"INVERT",docode,0,invn-invf
.invf   ld a,b
        cpl
        ld b,a
        ld a,c
        cpl
        ld c,a
.invn
        next

;C NEGATE   x1 -- x2            two's complement
    head NEGATE,6,"NEGATE",docode,0,negtn-negtf
.negtf  ld a,b
        cpl
        ld b,a
        ld a,c
        cpl
        ld c,a
        inc bc
.negtn
        next

;Z POLARITY	n -- 1 | 0 | -1		sign of n (1=+ve,-1=-ve,0=zero)
    head POLARITY,8,"POLARITY",docode,0,polar1-polarf
.polarf ld a,b
	or c
	jr z,polar1		; leave BC=0 if it is already
	rl b			; set carry if BC -ve
	sbc hl,hl		; HL=0 if BC +ve, or -1 if BC -ve
	ld b,h
	ld a,l
	or 1
	ld c,a			; TOS=1 or -1
.polar1
        next

;C 1+      n1/u1 -- n2/u2       add 1 to TOS
    head ONEPLUS,2,"1+",docode,0,onepln-oneplf
.oneplf inc bc
.onepln
        next

;C 1-      n1/u1 -- n2/u2     subtract 1 from TOS
    head ONEMINUS,2,"1-",docode,0,onemin-onemif
.onemif dec bc
.onemin
        next

;Z ><      x1 -- x2         swap bytes (not ANSI)
    head SWAPBYTES,2,"><",docode,0,swapbn-swapbf
.swapbf ld a,b
        ld b,c
        ld c,a
.swapbn
        next

;Z SPLIT        x -- n1 n2      split word into bytes (n1=low)
    head SPLIT,5,"SPLIT",docode,0,splitn-splitf
.splitf ld l,c
        ld h,0
        push hl         ; 2OS=low byte
        ld c,b
        ld b,h          ; TOS=high byte
.splitn
        next

;Z JOIN         n1 n2 -- x      join bytes to form word (n1=low)
    head JOIN,4,"JOIN",docode,0,joinn-joinf
.joinf  pop hl
        ld b,c          ; high byte=TOS
        ld c,l          ; low byte=2OS
.joinn
        next

;Z NSPLIT       x n -- x1 x2    split word to n high bits (x2), rest low (x1)
    head NSPLIT,6,"NSPLIT",docode,0,0
        ld a,c          ; A=#bits to go in high part
        pop hl
        ld b,h
        ld c,l
        push de         ; save IP
        ld de,0         ; mask
        push af
        sub 16
        neg             ; A=#bits to go in low part
        and a
        jr z,nsplit2
.nsplit1 scf
        rl e            ; build mask
        rl d
        dec a
        jr nz,nsplit1
.nsplit2 ld a,h
        and d           ; mask low bits
        ld h,a
        ld a,l
        and e
        ld l,a          ; HL=low part (x1)
        pop af
        and a
        jr z,nsplit4
.nsplit3 push af
        ld a,b
        rlca
        rl c            ; rotate bits
        rl b
        ld a,d
        rlca
        rl e            ; rotate mask
        rl d
        pop af
        dec a
        jr nz,nsplit3
.nsplit4 ld a,d
        cpl
        and b           ; mask high bits
        ld b,a
        ld a,e
        cpl
        and c
        ld c,a
        pop de          ; restore IP
        push hl         ; 2OS=low part
        next

;Z NJOIN        x1 x2 n -- x            Join n high bits (x1) with low bits
    head NJOIN,5,"NJOIN",docode,0,0
        ld a,c          ; A=#bits contained in x2
        pop hl          ; HL=x2
        pop bc          ; BC=x1
        push de         ; save IP
        ld de,0         ; mask
        and a
        jr z,njoin2
.njoin1 push af
        ld a,l
        rrca
        rr h
        rr l            ; shift n bits to high part of HL
        scf
        rr d            ; build mask
        rr e
        pop af
        dec a
        jr nz,njoin1
.njoin2 ld a,h
        and d           ; mask high bits
        ld h,a
        ld a,l
        and e
        ld l,a
        ld a,d
        cpl
        and b
        or h            ; combine low bits
        ld b,a
        ld a,e
        cpl
        and c
        or l
        ld c,a
        pop de          ; restore IP
        next
        
;C 2*      x1 -- x2         arithmetic left shift
    head TWOSTAR,2,"2*",docode,0,tstarn-tstarf
.tstarf sla c
        rl b
.tstarn
        next

;C 2/      x1 -- x2        arithmetic right shift
    head TWOSLASH,2,"2/",docode,0,tslasn-tslasf
.tslasf sra b
        rr c
.tslasn
        next

;C LSHIFT  x1 u -- x2    logical L shift u places
    head LSHIFT,6,"LSHIFT",docode,0,lshifn-lshiff
.lshiff ld b,c        ; b = loop counter
        pop hl        ;   NB: hi 8 bits ignored!
        inc b         ; test for counter=0 case
        jr lsh2
.lsh1   add hl,hl     ; left shift HL, n times
.lsh2   djnz lsh1
        ld b,h        ; result is new TOS
        ld c,l
.lshifn
        next

;C RSHIFT  x1 u -- x2    logical R shift u places
    head RSHIFT,6,"RSHIFT",docode,0,rshifn-rshiff
.rshiff ld b,c        ; b = loop counter
        pop hl        ;   NB: hi 8 bits ignored!
        inc b         ; test for counter=0 case
        jr rsh2
.rsh1   srl h         ; right shift HL, n times
        rr l
.rsh2   djnz rsh1
        ld b,h        ; result is new TOS
        ld c,l
.rshifn
        next

;C +!     n/u a-addr --       add cell to memory
    head PLUSSTORE,2,"+!",docode,0,plstn-plstf
.plstf  pop hl
        ld a,(bc)       ; low byte
        add a,l
        ld (bc),a
        inc bc
        ld a,(bc)       ; high byte
        adc a,h
        ld (bc),a
        pop bc          ; pop new TOS
.plstn
        next

; COMMON CONSTANT VALUES ========================

;Z 0	-- 0		push 0 onto the stack
    head ZERO,1,"0",docode,0,zeron-zerof
.zerof  push bc
.tosfalse
	ld bc,0
.zeron
        next

;Z 1    -- 1		push 1 onto the stack
    head ONE,1,"1",docode,0,onen-onef
.onef   push bc
	ld bc,1
.onen
        next

;Z -1	-- -1		push -1 onto the stack
    head MINUSONE,2,"-1",docode,0,monen-monef
.monef  push bc
	ld bc,-1
.monen
        next

; COMPARISON OPERATIONS =========================

;C 0=     n/u -- flag    return true if TOS=0
    head ZEROEQUAL,2,"0=",docode,0,zeqln-zeqlf
.zeqlf  ld a,b
        or c            ; result=0 if bc was 0
        sub 1           ; cy set   if bc was 0
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
.zeqln
        next

;C 0<     n -- flag      true if TOS negative
    head ZEROLESS,2,"0<",docode,0,zlessn-zlessf
.zlessf sla b           ; sign bit -> cy flag
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
.zlessn
        next

;X 0<>	n/u -- flag	return true if TOS<>0
    head ZERONOTEQUAL,3,"0<>",docode,0,zneqln-zneqlf
.zneqlf ld a,b
	or c		; A=0 if BC=0
	sub 1		; set carry if BC=0
	sbc a,a		; A=-1 if BC=0, 0 otherwise
	cpl		; A=0 if BC=0, -1 otherwise
	ld b,a		; move to TOS
	ld c,a
.zneqln
        next

;X 0>	n -- flag	true if TOS positive
    head ZEROGREATER,2,"0>",docode,0,zeroln-zerolf
.zerolf	bit 7,b		; test sign bit
	jr z,zerol2	; if positive, go to test <>0
	ld bc,0		; if negative, make=0 so next test is false
.zerol2	ld a,b
	or c
	sub 1
	sbc a,a
	cpl
	ld b,a
	ld c,a
.zeroln
	next

;C =      x1 x2 -- flag         test x1=x2
    head EQUAL,1,"=",docode,0,equaln-equalf
.equalf pop hl
        or a
        sbc hl,bc       ; x1-x2 in HL, SZVC valid
	ld bc,0
        jr nz,equaln	; false if different
        dec bc          ; true if equal
.equaln
        next

;X <>     x1 x2 -- flag    test not eq (not ANSI)
    head NOTEQUAL,2,"<>",docode,0,nequln-nequlf
.nequlf pop hl
	or a
	sbc hl,bc
	ld bc,0
	jr z,nequln	; false if equal
	dec bc		; true if not equal
.nequln
        next

;C <      n1 n2 -- flag        test n1<n2, signed
    head LESS,1,"<",docode,0,0
        pop hl
        or a
        sbc hl,bc       ; n1-n2 in HL, SZVC valid
; if result negative & not OV, n1<n2
; neg. & OV => n1 +ve, n2 -ve, rslt -ve, so n1>n2
; if result positive & not OV, n1>=n2
; pos. & OV => n1 -ve, n2 +ve, rslt +ve, so n1<n2
; thus OV reverses the sense of the sign bit
        jp pe,revsense  ; if OV, use rev. sense
        jp p,tosfalse   ;   if +ve, result false
.tostrue ld bc,$ffff   ;   if -ve, result true
        next
.revsense jp m,tosfalse ; OV: if -ve, reslt false
        jr tostrue      ;     if +ve, result true

;C >     n1 n2 -- flag         test n1>n2, signed
    head GREATER,1,">",docolon,0,0
        defw SWOP,LESS,EXIT

;C U<    u1 u2 -- flag       test u1<n2, unsigned
    head ULESS,2,"U<",docode,0,ulessn-ulessf
.ulessf pop hl
        or a
        sbc hl,bc       ; u1-u2 in HL, SZVC valid
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
.ulessn
        next

;X DU<	ud1 ud2 -- flag		test ud1<ud2, unsigned
    head DULESS,3,"DU<",docode,0,dulesn-dulesf
.dulesf pop hl
	ex (sp),hl	; HL=low word of ud1, BC=low word of ud2
	and a
	sbc hl,bc	; subtract low word
	pop bc
	pop hl
	sbc hl,bc	; subtract high word with carry
	sbc a,a		; propogate carry through A
	ld b,a		; put 0000 or FFFF in TOS
	ld c,a
.dulesn
        next

;X U>    u1 u2 -- flag     u1>u2 unsgd (not ANSI)
    head UGREATER,2,"U>",docode,0,ugrtn-ugrtf
.ugrtf	ld h,b
	ld l,c
	pop bc
	or a
	sbc hl,bc	; u2-u1 in HL, SZVC valid
	sbc a,a		; propagate carry through A
	ld b,a
	ld c,a
.ugrtn
	next

; LOOP AND BRANCH OPERATIONS ====================

;Z branch   --                  branch always
    nohead BRANCH,docode
.dobranch ld a,(de)     ; get inline value => IP
        ld l,a
        inc de
        ld a,(de)
        ld h,a
        nexthl

;Z ?branch   x --              branch if TOS zero
    nohead QBRANCH,docode
        ld a,b
        or c            ; test old TOS
        pop bc          ; pop new TOS
        jr z,dobranch   ; if old TOS=0, branch
        inc de          ; else skip inline value
        inc de
        next

;Z (do)    n1|u1 n2|u2 --  R: -- sys1 sys2
;Z                          run-time code for DO
; '83 and ANSI standard loops terminate when the
; boundary of limit-1 and limit is crossed, in
; either direction.  This can be conveniently
; implemented by making the limit 8000h, so that
; arithmetic overflow logic can detect crossing.
; I learned this trick from Laxen & Perry F83.
; fudge factor = 8000h-limit, to be added to
; the start value.
    nohead XDO,docode
        ex de,hl
        ex (sp),hl   ; IP on stack, limit in HL
        ex de,hl
        ld hl,$8000
        or a
        sbc hl,de    ; 8000-limit in HL
        dec ix       ; push this fudge factor
        ld (ix+0),h  ;    onto return stack
        dec ix       ;    for later use by 'I'
        ld (ix+0),l
        add hl,bc    ; add fudge to start value
        dec ix       ; push adjusted start value
        ld (ix+0),h  ;    onto return stack
        dec ix       ;    as the loop index.
        ld (ix+0),l
        pop de       ; restore the saved IP
        pop bc       ; pop new TOS
        next

;Z (loop)   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for LOOP
; Add 1 to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.  Note that LOOP
; terminates when index=8000h.
    nohead XLOOP,docode
        exx
        ld bc,1
.looptst ld l,(ix+0)  ; get the loop index
        ld h,(ix+1)
        or a
        adc hl,bc    ; increment w/overflow test
        jp pe,loopterm  ; overflow=loop done
        ; continue the loop
        ld (ix+0),l  ; save the updated index
        ld (ix+1),h
        exx
        jr dobranch  ; take the inline branch
.loopterm ; terminate the loop
        ld bc,4      ; discard the loop info
        add ix,bc
        exx
        inc de       ; skip the inline branch
        inc de
        next

;Z (+loop)   n --   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for +LOOP
; Add n to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.
    nohead XPLUSLOOP,docode
        pop hl      ; this will be the new TOS
        push bc
        ld b,h
        ld c,l
        exx
        pop bc      ; old TOS = loop increment
        jr looptst

;C I        -- n   R: sys1 sys2 -- sys1 sys2
;C                  get the innermost loop index
    head II,1,"I",docode,0,0
        push bc     ; push old TOS
        ld l,(ix+0) ; get current loop index
        ld h,(ix+1)
        ld c,(ix+2) ; get fudge factor
        ld b,(ix+3)
        or a
        sbc hl,bc   ; subtract fudge factor,
        ld b,h      ;   returning true index
        ld c,l
        next

;C J        -- n   R: 4*sys -- 4*sys
;C                  get the second loop index
    head JJ,1,"J",docode,0,0
        push bc     ; push old TOS
        ld l,(ix+4) ; get current loop index
        ld h,(ix+5)
        ld c,(ix+6) ; get fudge factor
        ld b,(ix+7)
        or a
        sbc hl,bc   ; subtract fudge factor,
        ld b,h      ;   returning true index
        ld c,l
        next

;C UNLOOP   --   R: sys1 sys2 --  drop loop parms
    head UNLOOP,6,"UNLOOP",docode,0,0
        inc ix
        inc ix
        inc ix
        inc ix
        next

; MULTIPLY AND DIVIDE ===========================

;C UM*     u1 u2 -- ud   unsigned 16x16->32 mult.
    head UMSTAR,3,"UM*",docode,0,0
        push bc
        exx
        pop bc      ; u2 in BC
        pop de      ; u1 in DE
        ld hl,0     ; result will be in HLDE
        ld a,17     ; loop counter
        or a        ; clear cy
.umloop rr h
        rr l
        rr d
        rr e
        jr nc,noadd
        add hl,bc
.noadd  dec a
        jr nz,umloop
        push de     ; lo result
        push hl     ; hi result
        exx
        pop bc      ; put TOS back in BC
        next

;C UM/MOD   ud u1 -- u2 u3   unsigned 32/16->16
    head UMSLASHMOD,6,"UM/MOD",docode,0,0
        push bc
        exx
        pop bc      ; BC = divisor
        pop hl      ; HLDE = dividend
        pop de
        ld a,16     ; loop counter
        sla e
        rl d        ; hi bit DE -> carry
.udloop adc hl,hl   ; rot left w/ carry
        jr nc,udiv3
        ; case 1: 17 bit, cy:HL = 1xxxx
        or a        ; we know we can subtract
        sbc hl,bc
        or a        ; clear cy to indicate sub ok
        jr udiv4
        ; case 2: 16 bit, cy:HL = 0xxxx
.udiv3  sbc hl,bc   ; try the subtract
        jr nc,udiv4 ; if no cy, subtract ok
        add hl,bc   ; else cancel the subtract
        scf         ;   and set cy to indicate
.udiv4  rl e        ; rotate result bit into DE,
        rl d        ; and next bit of DE into cy
        dec a
        jr nz,udloop
        ; now have complemented quotient in DE,
        ; and remainder in HL
        ld a,d
        cpl
        ld b,a
        ld a,e
        cpl
        ld c,a
        push hl     ; push remainder
        push bc
        exx
        pop bc      ; quotient remains in TOS
        next

; BLOCK AND STRING OPERATIONS ===================

;C FILL   c-addr u char --  fill memory with char
    head FILL,4,"FILL",docode,0,filln-fillf
.fillf  ld a,c          ; character in a
        exx             ; use alt. register set
        pop bc          ; count in bc
        pop de          ; address in de
        or a            ; clear carry flag
        ld hl,$ffff
        adc hl,bc       ; test for count=0 or 1
        jr nc,filldone  ;   no cy: count=0, skip
        ld (de),a       ; fill first byte
        jr z,filldone   ;   zero, count=1, done
        dec bc          ; else adjust count,
        ld h,d          ;   let hl = start adrs,
        ld l,e
        inc de          ;   let de = start adrs+1
        ldir            ;   copy (hl)->(de)
.filldone exx           ; back to main reg set
        pop bc          ; pop new TOS
.filln
        next

;X CMOVE   c-addr1 c-addr2 u --  move from bottom
; as defined in the ANSI optional String word set
; On byte machines, CMOVE and CMOVE> are logical
; factors of MOVE.  They are easy to implement on
; CPUs which have a block-move instruction.
    head CMOVE,5,"CMOVE",docode,0,cmoven-cmovef
.cmovef push bc
        exx
        pop bc      ; count
        pop de      ; destination adrs
        pop hl      ; source adrs
        ld a,b      ; test for count=0
        or c
        jr z,cmovedone
        ldir        ; move from bottom to top
.cmovedone exx
        pop bc      ; pop new TOS
.cmoven
        next

;X CMOVE>  c-addr1 c-addr2 u --  move from top
; as defined in the ANSI optional String word set
    head CMOVEUP,6,"CMOVE>",docode,0,cmovun-cmovuf
.cmovuf push bc
        exx
        pop bc      ; count
        pop hl      ; destination adrs
        pop de      ; source adrs
        ld a,b      ; test for count=0
        or c
        jr z,umovedone
        add hl,bc   ; last byte in destination
        dec hl
        ex de,hl
        add hl,bc   ; last byte in source
        dec hl
        lddr        ; move from top to bottom
.umovedone exx
        pop bc      ; pop new TOS
.cmovun
        next


;Z SCAN    c-addr u c -- c-addr' u'
;Z                      find matching char
;  If scanning for BLs, all control characters match too
    head SCAN,4,"SCAN",docode,0,0
        ld a,c      ; scan character
        exx
        pop bc      ; count
        pop hl      ; address
        ld e,a      ; test for count=0
        ld a,b
        or c
        jr z,scandone
        ld a,e
	cp $20		; are we scanning for blanks+control chars?
	jr z,scanbllp
        cpir        ; scan 'til match or count=0
        jr nz,scandone  ; no match, BC & HL ok
        inc bc          ; match!  undo last to
        dec hl          ;   point at match char
.scandone push hl   ; updated address
        push bc     ; updated count
        exx
        pop bc      ; TOS in bc
        next
.scanbllp ld a,$20
	cp (hl)
	jr nc,scandone	; done if character<=20h found
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,scanbllp	; loop back for more chars
	jr scandone


;Z S=    c-addr1 c-addr2 u -- n   string compare
;Z             n<0: s1<s2, n=0: s1=s2, n>0: s1>s2
    head SEQUAL,2,"S=",docode,0,0
        push bc
        exx
        pop bc      ; count
        pop hl      ; addr2
        pop de      ; addr1
        ld a,b      ; test for count=0
        or c
        jr z,smatch     ; by definition, match!
.sloop  ld a,(de)
        inc de
        cpi
        jr nz,sdiff     ; char mismatch: exit
        jp pe,sloop     ; count not exhausted
.smatch ; count exhausted & no mismatch found
        exx
        ld bc,0         ; bc=0000  (s1=s2)
        jr snext
.sdiff  ; mismatch!  undo last 'cpi' increment
        dec hl          ; point at mismatch char
        cp (hl)         ; set cy if char1 < char2
        sbc a,a         ; propagate cy thru A
        exx
        ld b,a          ; bc=FFFF if cy (s1<s2)
        or 1            ; bc=0001 if ncy (s1>s2)
        ld c,a
.snext
        next


;X SEARCH	caddr1 u1 caddr2 u2 -- caddr3 u3 flag	substring search
    head SEARCH,6,"SEARCH",docode,0,0
	pop hl		; HL/BC=substring
	ld a,b
	or c
	jr z,search3	; if zero length substring, immediate match
	exx
	pop bc
	pop hl
	push hl
	push bc		; HL'/BC'=main string
.search1 ld a,b
	or c
	jr z,search5	; no match if end of main string
	exx
	ld a,(hl)
	exx
	cpir		; search for first substring char
	jr nz,search5	; no match if not found
	dec hl		; backup to matching char
	inc bc
	push bc
	exx
	ex (sp),hl
	and a
	sbc hl,bc	; test length left against substring length
	pop hl
	exx
	jr c,search5	; no match if not long enough
	push hl
	exx
	ex de,hl
	ex (sp),hl	; save IP, get HL=main string, DE=substring
	push de		; save substring
	push bc
.search2 ld a,(de)
	cp (hl)
	jr nz,search4	; stop trying if char mismatch
	inc de
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,search2	; for rest of substring
	pop de
	pop de
	pop de		; discard substring, restore IP
	pop bc
	pop bc		; discard caddr1/u1
	exx
	push hl		; stack caddr3/u3
	push bc
	exx
.search3 ld bc,-1       ; true flag
	jr search6
.search4 pop bc         ; restore substring
	pop hl
	pop de		; restore IP
	exx
	inc hl		; move past first matching char
	dec bc
	jr search1	; back for another try
.search5 exx
	ld bc,0		; false flag
.search6
        next

;Z >UPPER       c-addr u --     Convert string to uppercase
     head TOUPPER,6,">UPPER",docode,0,tupprn-tupprf
.tupprf pop hl          ; HL=address, BC=count
.tupp0  ld a,b
        or c
        jr z,tupp2      ; end of string reached
        ld a,(hl)
        cp 'a'
        jr c,tupp1
        cp '{'
        jr nc,tupp1
        and 223         ; convert lowercase letters only
        ld (hl),a
.tupp1  inc hl
        dec bc
        jr tupp0
.tupp2  pop bc          ; get new TOS
.tupprn
        next

;Z >LOWER       c-addr u --     Convert string to lowercase
     head TOLOWER,6,">LOWER",docode,0,tlowrn-tlowrf
.tlowrf pop hl          ; HL=address, BC=count
.tlow0  ld a,b
        or c
        jr z,tlow2      ; end of string reached
        ld a,(hl)
        cp 'A'
        jr c,tlow1
        cp '['
        jr nc,tlow1
        or 32           ; convert uppercase letters only
        ld (hl),a
.tlow1  inc hl
        dec bc
        jr tlow0
.tlow2  pop bc
.tlowrn
        next




