; Z88 CamelForth
; camellow.s (Primitives for Z88)

	include "#stdio.def"
	include "#fileio.def"
	include	"#memory.def"
	include "#error.def"
	include	"#saverst.def"
	include "#time.def"
	include "#integer.def"
	include "#director.def"
	include "#syspar.def"
	include	"#dor.def"
	include	"#map.def"
	include	"#screen.def"

; Initial stack pointer

;Z S0       -- a-addr     end of parameter stack
    nohead S0,docode
	push bc
	ld bc,(initsp)
	next

; Initial return stack pointer (per task)

;Z R0			-- a-addr	End of return stack
    nohead R0,docode
	push bc
	ld c,(iy+usr_r0)
	ld b,(iy+usr_r0+1)
	next

; Bad application memory to give back to system

;Z HOLE			-- a-addr	Hole to give back (2 cells)
    head HOLE,4,"HOLE",docon,0,0
	defw usr_hole

; Deferred words

;Z (RC_QUIT)	--	RC_QUIT action
;   DEFER (RC_QUIT)
    head BRRCQUIT,9,"(RC_QUIT)",dodefer,0,0
	defw usr_quit

;Z (RC_DRAW)	--	RC_DRAW action
;   DEFER (RC_DRAW)
    head BRRCDRAW,9,"(RC_DRAW)",dodefer,0,0
	defw usr_draw

;Z (RC_ESC)	--	RC_ESC action
;   DEFER (RC_ESC)
    head BRRCESC,8,"(RC_ESC)",dodefer,0,0
	defw usr_esc

;Z (KEY)	--	KEY special keys action
;   DEFER (KEY)
    head BRKEY,5,"(KEY)",dodefer,0,0
	defw usr_brkey

;Z CAMEL		--		Cold start for CamelForth
;   CINIT ." Z88 CamelForth vX.YZ"
;   FALSE GBOUND
;   QUIT ;
    head CAMEL,5,"CAMEL",docolon,0,0
	defw CINIT
	defw LIT,versionmsg,LIT,endversionmsg-versionmsg,TYPE
	defw FALSE,GBOUND,QUIT,EXIT

;Z OS_ESC		reason -- results	ESCape handling
   head OSESC,6,"OS_ESC",docode,0,ozescn-ozescf
.ozescf ld a,c          ; A=reason
	call_oz(os_esc)
	push af
	pop bc		; TOS=results
.ozescn
	next

;Z ESCAPE		--		ESCAPE standard RC_ESC vector
;   1 OS_ESC DROP TRUE ABORT" Escape" ;
.escheader
    head ESCAPE,6,"ESCAPE",docolon,0,0
	defw ONE,OSESC,DROP
	defw TRUE,LIT,escheader+2	; use word's own name as message
	defw COUNT,QABORT


; Subroutine used to test error code in A and execute any appropriate
; vectored word

.vector ret nc          ; exit if no error
	push af		; save error code
	push de		; save IP
	push bc		; save TOS
	push hl		; save W
	ld de,retrn
        ld hl,(usr_quit)
	cp rc_quit
	jr z,dovect
        ld hl,(usr_draw)
	cp rc_draw
	jr nz,norcdr
	xor a
	ld (basegr+2),a	; disable graphics on RC_DRAW
	jr dovect
.norcdr ld hl,(usr_esc)
	cp rc_esc
	jr nz,novect	; if none of these errors, don't execute a vector
.dovect jp (hl)         ; execute vector
.retrn  defw novect       ; re-enter code
.novect pop hl          ; restore W
	pop bc		; restore TOS
	pop de		; restore IP
	pop af		; restore error code
	ret


; GENERAL-PURPOSE PRIMITIVES

;Z 0PAD		-- addr		Null-terminated string scratch area
    head ZEROPAD,4,"0PAD",docon,0,0
        defw usr_zpad

;Z S>0   c-addr n -- c-addr'   Convert string to null-terminated
    head STRTONUL,3,"S>0",docode,0,strnun-strnuf
.strnuf pop hl          ; HL=source address
	ld b,0		; ensure count < 256 to avoid memory corruption
	push de		; save IP
	ld de,usr_zpad
	ld a,c
	and a
	jr z,STRTONUL2	; skip copy if zero length
	ldir		; copy
.STRTONUL2
	ld a,c
	ld (de),a	; add nul
	pop de
	ld bc,usr_zpad	; TOS=address of null-terminated string (0PAD)
.strnun
	next


; MAILBOXING/NULL-TERMINATED STRINGS

;Z (0")		-- c-addr	run-time action of 0"
;   R> COUNT OVER SWAP + ALIGNED >R ;
    nohead BRZQUOTE,docolon
	defw RFROM,COUNT,OVER,SWOP,PLUS,TOR,EXIT

;Z 0"		"ccc<quote>" --	compile null-terminated string
;   POSTPONE (0")  HERE (")
;   DUP C@ 1+ SWAP C!
;   0 C, ; IMMEDIATE
    head ZQUOTE,2,"0"&$22,docolon,1,0
	defw LIT,BRZQUOTE,COMMAXT,HERE,BRQUOTE
	defw DUP,CFETCH,ONEPLUS,SWOP,CSTORE
	defw ZERO,CCOMMA,EXIT

;Z 0TYPE        c-addr --       Type null-terminated string
    head ZTYPE,5,"0TYPE",docode,0,ztypn-ztypf
.ztypf  ld h,b
        ld l,c
        call_oz(gn_sop)
        pop bc
.ztypn
	next

;Z CHECKMAIL	c-addr -- c-addr n true | false		Check for mail
    head CHECKMAIL,9,"CHECKMAIL",docode,0,endmail-chkmail
.chkmail push de                ; save IP
	ld d,b
	ld e,c
	ld hl,usr_zpad
	ld bc,255
	ld a,4		; SR_RPD, read mailbox data
	call_oz(os_sr)
	pop de		; restore IP
	jr c,nomail	; move on if mail not found
	push hl		; 3OS=buffer address
	push bc		; 2OS=data length
	ld bc,-1	; true, got mail
	jr endmail
.nomail ld bc,0         ; false, no mail
.endmail
	next

;Z SENDMAIL	c-addr1|0 c-addr2 n --	Send data to mailbox
    head SENDMAIL,8,"SENDMAIL",docode,0,smailn-smailf
.smailf pop hl          ; HL=data address
	ex de,hl
	ex (sp),hl
	ex de,hl	; save IP, DE=mailtype
	ld a,3		; SR_WPD, write mailbox data
	call_oz(os_sr)
	pop de		; restore IP
	pop bc		; get new TOS
.smailn
	next

; CONSOLE I/O =============================================================


;Z CEMIT     c --    output character to standard output
    head CEMIT,5,"CEMIT",docode,0,cemtn-cemtf
.cemtf  ld a,c          ; A=TOS
	pop bc		; get new TOS
	call_oz(os_out)
.cemtn
	next

;Z CTYPE     c-addr +n --	type line to console
    head CTYPE,5,"CTYPE",docode,0,ctypen-ctypef
.ctypef pop hl          ; HL=address
	jr ctyp2
.ctyp1  ld a,(hl)       ; get next char
	inc hl
	call_oz(os_out)
	dec bc
.ctyp2  ld a,b
	or c
	jr nz,ctyp1
	pop bc		; get new TOS
.ctypen
	next

;X EMIT?   -- f    test if output device ready
    head EMITQUERY,5,"EMIT?",docon,0,0
	defw -1		; always true


;X PAGE		start new page of output
    head PAGE,4,"PAGE",docolon,0,0
	defw LIT,12,EMIT,EXIT


;X AT-XY   x y --    move cursor to (x,y)
;   1 EMIT [CHAR] 3 EMIT [CHAR] @ EMIT
;   SWAP 32 + EMIT 32 + EMIT ;
    head ATXY,5,"AT-XY",docolon,0,0
	defw ONE,EMIT,LIT,51,EMIT,LIT,64,EMIT
	defw SWOP,LIT,32,PLUS,EMIT,LIT,32,PLUS,EMIT,EXIT
	

;X EKEY     -- u    get event code from standard input (deferred)
;   DEFER EKEY
    head EKEY,4,"EKEY",dodefer,0,0
	defw usr_ekey


;Z EKEYM	    -- u    get event code from standard input (multi version)
    nohead EKEYM,docolon
	defw mkey2
.mkey2	push bc		; move TOS to 2OS
	call mrkey
	and a
	jr z,mkyevent	; move on if we have a key event code
	ld b,0		; just an ordinary character
	cp $7f
	jr nz,mkeyend
	ld a,$08	; substitute BS for Z88 DEL
	jr mkeyend
.mkyevent call mrkey    ; get another code
	ld b,1		; special events have codes 100h upwards
.mkeyend ld c,a         ; TOS=char/code
	jp EXIT

; Subroutine used to interrogate OZ for a key code
.mrkey  ld de,mrkey2
	jp PAUSE	; allow other tasks to run
.mrkey2	defw mrkey3
.mrkey3	ld bc,1		; only wait 1/100s for a key
	call_oz(os_tin)
	ret nc
	cp rc_time
	jr z,mrkey	; if it was a timeout, loop back
	scf
	call vector	; execute appropriate error vector word
	pop hl		; discard return address
	ld bc,256
	jp EXIT

;Z EKEYS	    -- u    get event code from standard input (single version)
    nohead EKEYS,docode
	push bc		; move TOS to 2OS
	call srkey
	and a
	jr z,skyevent	; move on if we have a key event code
	ld b,0		; just an ordinary character
	cp $7f
	jr nz,skeyend
	ld a,$08	; substitute BS for Z88 DEL
	jr skeyend
.skyevent call srkey    ; get another code
	ld b,1		; special events have codes 100h upwards
.skeyend ld c,a         ; TOS=char/code
	next

; Subroutine used to interrogate OZ for a key code
.srkey  call_oz(os_in)
	ret nc
	call vector	; execute appropriate error vector word
	pop hl		; discard return address
	ld bc,256
	next
	
;X EKEY?     -- f    return true if event waiting
    head EKEYQUERY,5,"EKEY?",docode,0,ekeyqn-ekeyqf
.ekeyqf push bc         ; move TOS to 2OS
	call_oz(os_xin)
	ccf
	sbc a,a		; A=0 if input not available, -1 if may be 
	ld c,a
	ld b,a		; TOS=flag
.ekeyqn
	next


;X EKEY>CHAR  u -- u false | c true     convert key event to char
;   DUP 256 < ;
    head EKEYTOCHAR,9,"EKEY>CHAR",docolon,0,0
	defw DUP,LIT,256,LESS,EXIT


;Z PENDING-CHAR  -- addr    User variable holding a pending key character
    nohead PENDINGCHAR,docon
	defw usr_pending


;X KEY?  -- f	 return true if char waiting
;   PENDING-CHAR @ 0< IF  BEGIN  EKEY?  WHILE  EKEY EKEY>CHAR
;   IF  PENDING-CHAR ! TRUE EXIT  THEN  (KEY)  REPEAT
;   FALSE EXIT  THEN  TRUE ;
    head KEYQUERY,4,"KEY?",docolon,0,0
	defw PENDINGCHAR,FETCH,ZEROLESS,QBRANCH,KEYQY5
.KEYQY2 defw EKEYQUERY,QBRANCH,KEYQY4,EKEY,EKEYTOCHAR
	defw QBRANCH,KEYQY3,PENDINGCHAR,STORE,TRUE,EXIT
.KEYQY3 defw BRKEY,BRANCH,KEYQY2
.KEYQY4 defw FALSE,EXIT
.KEYQY5 defw TRUE,EXIT


;C CKEY  -- c    get keyboard character
;   PENDING-CHAR @ 0< IF  BEGIN  EKEY EKEY>CHAR 0=  WHILE  (KEY)   REPEAT
;   EXIT  THEN  PENDING-CHAR @  -1 PENDING-CHAR ! ;
    head CKEY,4,"CKEY",docolon,0,0
	defw PENDINGCHAR,FETCH,ZEROLESS,QBRANCH,KEY4
.KEY2   defw EKEY,EKEYTOCHAR,ZEROEQUAL,QBRANCH,KEY3,BRKEY
	defw BRANCH,KEY2
.KEY3   defw EXIT
.KEY4   defw PENDINGCHAR,FETCH,MINUSONE,PENDINGCHAR,STORE,EXIT


;X MS     u --		wait at least u milliseconds
    head MS,2,"MS",docolon,0,0
	defw PAUSE	; allow other tasks to run
.stMS	defw stMS+2
	ld l,10		; only run other tasks every 10ms
.reMS   ld a,b
	or c
	jr z,endMS
	push bc
	ld bc,133
.reMS2  dec bc
	ld a,b
	or c
	jr nz,reMS2
	pop bc
	dec bc
	dec l
	jr nz,reMS
	ld de,stMS
	jp PAUSE	; allow other tasks to run
.endMS  pop bc          ; get new TOS
	jp EXIT

;Z INTT&D ( -- addr )		 get machine time & date (internal)
    head INTTD,6,"INTT&D",docode,0,inttdn-inttdf
.inttdf push bc         ; save TOS
	push de		; save IP
	ld de,usr_zpad+3
	call_oz(gn_gmd)
	dec de
	ld a,(de)
	ld c,a		; C=low byte of machine date
	ld de,usr_zpad
	call_oz(gn_gmt)
	pop de		; restore IP
	ld bc,usr_zpad	; TOS=address
.inttdn
	next


;Z INT>T&D	addr -- s min h d mth y
    head INTTOTD,7,"INT>T&D",docode,0,0
	ld (ix-1),e
	ld (ix-2),d	; save IP above top of return stack
	ld h,b
	ld l,c		; HL=address of internal time
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld b,(hl)
	inc hl
	ld (ix-3),l
	ld (ix-4),h	; save address of internal date
	ex de,hl	; BHL=internal time
	ld c,0
	ld de,100
	call_oz(gn_d24)
	ld e,60
	call_oz(gn_d24)
	push de		; stack seconds
	ld e,60
	call_oz(gn_d24)
	push de		; stack minutes
	push hl		; stack hours
	ld h,(ix-4)
	ld l,(ix-3)
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	ld a,(hl)	; ABC=internal date
	call_oz(gn_die)
	ld a,c
	and 31
	ld c,a
	ld a,b
	ld b,0
	push bc		; stack day
	ld c,a
	push bc		; stack month
	ld b,d
	ld c,e		; TOS=year
	ld d,(ix-2)
	ld e,(ix-1)	; restore IP
	next


;X TIME&DATE ( -- s m h d m y )		 get current time & date
;   INTT&D INT>T&D ;
    head TIMEANDDATE,9,"TIME&DATE",docolon,0,0
	defw INTTD,INTTOTD,EXIT


;X BYE     i*x --    exit application
    head BYE,3,"BYE",docode,0,byen-byef
.byef   ld sp,(initsp)  ; restore initial SP
	pop bc		; and get rid of dummy value
	xor a		; no error to display
        ld hl,clients
        dec (hl)        ; remove our instance from #clients
        call_oz(os_bye)
.byen

; Subroutine to get the current cursor position and place it in
; HL. AF will be corrupted.

.getcursor      push bc
	push de
	ld bc,nq_wcur
	xor a
	call_oz(os_nq)			; get current cursor
	ld h,b
	ld l,c				; HL=cursor
	pop de
	pop bc
	ret

;Z AT?		-- x y
    head ATQ,3,"AT?",docode,0,atqn-atqf
.atqf   push bc         ; push TOS
	call getcursor
	ld c,h
	ld b,0		; TOS=y
	ld h,b
	push hl		; 2OS=x
.atqn
	next


; OZ FEATURES =============================================================

;Z *CLI  c-addr n --	Invoke the CLI
    head STARCLI,4,"*CLI",docolon,0,0
	defw SWOP,OVER,STRTONUL,dostarcli
.dostarcli
	ld h,b
	ld l,c		; HL=address
	pop bc
	ld b,0		; C=length of string
	push de		; save IP, just in case
	call_oz(dc_icl)
	pop de		; restore IP
	pop bc		; get new TOS
	jp EXIT


;Z *NAME  c-addr n --	Name the current instantiation of CamelForth
    head STARNAME,5,"*NAME",docolon,0,0
	defw STRTONUL,dostarname
.dostarname
	ld h,b
	ld l,c
	call_oz(dc_nam)
	pop bc		; get new TOS
	jp EXIT


; LOW-LEVEL FILE WORDS =================================================


;X BIN		fam1 -- fam2
;   ;
    head BIN,3,"BIN",docolon,1,0
	defw EXIT


;X R/O		     -- fam
;   1 CONSTANT R/O
    head RSLASHO,3,"R/O",docon,0,0
	defw 1


;X W/O		     -- fam
;   2 CONSTANT W/O
    head WSLASHO,3,"W/O",docon,0,0
	defw 2


;X R/W		     -- fam
;   3 CONSTANT R/W
    head RSLASHW,3,"R/W",docon,0,0
	defw 3


;X CREATE-FILE	caddr u fam -- fileid ior
;   >R 2DUP W/O (OPEN-FILE) ?DUP
;   IF  2SWAP 2DROP R> DROP
;   ELSE  CLOSE-FILE DROP R> OPEN-FILE  THEN ;
     head CREATEFILE,11,"CREATE-FILE",docolon,0,0
	defw TOR,TWODUP,WSLASHO,BROPENFILE,QDUP
	defw QBRANCH,cfile2
	defw TWOSWAP,TWODROP,RFROM,DROP,EXIT
.cfile2 defw CLOSEFILE,DROP,RFROM,OPENFILE
.cfile3 defw EXIT


;X OPEN-FILE	caddr u fam -- fileid ior
;   DUP W/O = IF  DROP R/W  THEN
;   (OPEN-FILE) ;
    head OPENFILE,9,"OPEN-FILE",docolon,0,0
	defw DUP,WSLASHO,EQUAL
	defw QBRANCH,opfil2
	defw DROP,RSLASHW
.opfil2 defw BROPENFILE,EXIT


;X DELETE-FILE	caddr u -- ior
    head DELETEFILE,11,"DELETE-FILE",docolon,0,0
	defw STRTONUL,dfile2
.dfile2 ld h,b
	ld l,c
	ld b,0
	call_oz(gn_del)
        jp setior       ; go to set result codes


;X FILE-POSITION	fileid -- ud ior
    head FILEPOSITION,13,"FILE-POSITION",docode,0,0
	ld a,1		; reason: FA_PTR
.osfrm  push de         ; save IP
	push bc
	ex (sp),ix	; save RP, IX=fileid
	ld de,0
	call_oz(os_frm)
	ex de,hl	; HLBC=result
	pop ix
	pop de		; restore RP & IP
	push bc
	push hl		; stack result
        jp setior2      ; go to set result codes


;X FILE-SIZE	fileid -- ud ior
    head FILESIZE,9,"FILE-SIZE",docode,0,0
	ld a,2		; reason: FA_EXT
	jr osfrm


;X READ-FILE	caddr u1 fileid -- u2 ior
    head READFILE,9,"READ-FILE",docode,0,0
	ex de,hl
	pop de
	ex (sp),hl	; BC=fileid, DE=u1, HL=caddr, IP saved on stack
	push bc
	ex (sp),ix	; IX=fileid, DE=u1, HL=caddr, IP & RP saved on stack
	ld b,d
	ld c,e		; BC=length
	push bc		; save length for later
	ld de,0
	ex de,hl	; HL=0, DE=caddr - move from file to memory
	call_oz(os_mv)
	pop hl		; HL=original length
	pop ix		; restore RP
	pop de		; and IP
	call vector	; execute any appropriate error vector
	push af		; save error code
	and a
	sbc hl,bc	; HL=length moved
	pop af
	push hl		; 2OS=length moved
	jr nc,setior2	; exit if no error
	cp 9
	jr z,setior2	; or if EOF (carry reset)
	scf
	jr setior2	; else set error


;X WRITE-FILE	caddr u1 fileid -- ior
    head WRITEFILE,10,"WRITE-FILE",docode,0,0
	ex de,hl
	pop de
	ex (sp),hl	; BC=fileid, DE=u1, HL=caddr, IP saved on stack
	push bc
	ex (sp),ix	; IX=fileid, DE=u1, HL=caddr, IP & RP saved on stack
	ld b,d
	ld c,e		; BC=length
	ld de,0		; DE=0, HL=caddr - move from memory to file
	call_oz(os_mv)
	pop ix		; restore RP
	pop de		; and IP
	call vector	; execute any appropriate error vector
	jr setior2	; go to set success/error


;Z (OPEN-FILE)	caddr u fam -- fileid ior
    nohead BROPENFILE,docolon
	defw TOR,STRTONUL,RFROM,openf2
.openf2 ld a,c          ; A=access-mode
	pop hl
	ld bc,255	; explicit name length 255, local filename
	push ix
	push de		; save RP & IP
	ld de,usr_zpad	; (place for expanded filename)
	call_oz(gn_opf)
	pop de
	ex (sp),ix	; restore RP & IP; 2OS=fileid
.setior ld bc,0         ; set ior=0 (success)
	jp nc,EXIT	; exit if no error
	ld b,$fe
	neg
	ld c,a		; set ior=-256-A
	jp EXIT


;X CLOSE-FILE	fileid -- ior
    head CLOSEFILE,10,"CLOSE-FILE",docode,0,exior-cfilef
.cfilef push bc
	ex (sp),ix	; save RP, get IX=fileid
	call_oz(gn_cl)
	pop ix		; restore RP
.setior2 ld bc,0        ; set ior=0 (success)
	jr nc,exior	; exit if no error
	ld b,$fe
	neg
	ld c,a		; set ior=-256-A
.exior
	next


;X READ-LINE	caddr u1 fileid -- u2 flag ior
    head READLINE,9,"READ-LINE",docode,0,0
        ex de,hl
        pop de
        ex (sp),hl
        push bc
        ex (sp),ix      ; IX=fileid, DE=u1, HL=caddr, IP & RP saved on stack
        push de         ; save max length for later
.rdllp  ld a,d
        or e
        jr z,endrl      ; move on if no more chars to read
        call_oz(os_gb)
        jr c,rlerr      ; if error, move on
        dec de
        ld (hl),a
        inc hl
        cp 13
        jr z,iseol      ; move on if EOL char (CR)
        cp 10
        jr nz,rdllp     ; loop back if not end-of-line (LF)
.iseol  inc de          ; ignore eol terminator in calculations
        pop hl
        sbc hl,de       ; HL=chars read
.rdlok  pop ix          ; restore RP
        pop de          ; restore IP
        push hl         ; 3OS=u2
        ld hl,-1
        push hl         ; 2OS=true
        ld bc,0
        jr exior        ; exit with no error
.endrl  pop hl          ; HL=chars to read=chars read
        jr rdlok
.rlerr  pop hl
        push af
        and a
        sbc hl,de       ; HL=actual chars read
        pop af
        pop ix          ; restore RP
        pop de          ; restore IP
        call vector     ; execute any appropriate error vector
        push hl         ; 3OS=u2
        cp 9
        jr z,rleof      ; move on if EOF error
.rlfail ld hl,0
        push hl         ; else flag=FALSE
        jr setior2
.rleof  ld a,h
        or l
        ld a,9
        jr nz,rlfail    ; fail if some chars were read
        push hl         ; special case: u2=0, flag=0 and ior=0
        ld b,h
        ld c,l
        jr exior


;X WRITE-LINE	caddr u1 fileid -- ior
    head WRITELINE,10,"WRITE-LINE",docode,0,0
	ex de,hl
	pop de
	ex (sp),hl
	push bc
	ex (sp),ix	; IX=fileid, DE=u1, HL=caddr, IP & RP saved on stack
	ld b,d
	ld c,e
	ld de,0
	call_oz(os_mv)
	jr c,wtlerr	; move on if error
	ld a,13
	call_oz(os_pb)
.wtlerr pop ix          ; restore RP
	pop de		; restore IP
	call vector	; execute any appropriate error vector
        jp setior2      ; go to set success/error


;X REPOSITION-FILE	ud fileid -- ior
    head REPOSITIONFILE,15,"REPOSITION-FILE",docode,0,0
	ld a,1		; reason=FA_PTR
.osfwm  pop hl
	ld (ix-1),h
	ld (ix-2),l
	pop hl
	ld (ix-3),h
	ld (ix-4),l	; place position on return stack
	push bc
	ex (sp),ix	; IX=fileid
	pop hl
	push hl		; save RP
	dec hl
	dec hl
	dec hl
	dec hl		; HL=pointer to 4-byte value
	call_oz(os_fwm)
	pop ix		; restore RP
        jp setior2      ; go to set success/error


;Z EXTEND-FILE		ud fileid -- ior
    head EXTENDFILE,11,"EXTEND-FILE",docode,0,0
	ld a,2		; reason=FA_EXT
	jr osfwm


;X RESIZE-FILE		ud fileid -- ior
;   DUP >R FILE-SIZE ?DUP
;   IF  R> DROP >R 2DROP 2DROP R> EXIT  THEN
;   2OVER 2SWAP DU<
;   IF  2DROP R> DROP -279 EXIT  THEN		; RC_FAIL if want to shorten
;   R> EXTEND-FILE ;
    head RESIZEFILE,11,"RESIZE-FILE",docolon,0,0
	defw DUP,TOR,FILESIZE,QDUP
	defw QBRANCH,resiz2
	defw RFROM,DROP,TOR,TWODROP,TWODROP,RFROM,EXIT
.resiz2 defw TWOOVER,TWOSWAP,DULESS
	defw QBRANCH,resiz3
	defw TWODROP,RFROM,DROP,LIT,-279,EXIT
.resiz3 defw RFROM,EXTENDFILE,EXIT


;X FILE-STATUS		caddr u -- x ior
;  For this Z88 implementation, x is an address of a data structure
;  holding creation & update time & dates, in internal machine format  
;  This may be overwritten on pre-emption or with use of other words
;  that access OZ, so should be moved elsewhere if required
;  Typically, you would immediately convert it to time & date on
;  the stack with INT>T&D
    head FILESTATUS,11,"FILE-STATUS",docolon,0,0
	defw LIT,6,BROPENFILE			; open DOR handle
	defw DUP,QBRANCH,fstat1,EXIT
.fstat1 defw fstat2
.fstat2 ex (sp),ix              ; save RP, get IX=DOR handle
	push de			; save IP
	ld de,usr_zpad
	ld bc,$4306		; creation time
	ld a,dr_rd
	call_oz(os_dor)
	ld de,usr_zpad+6
	ld bc,$5506		; update time
	ld a,dr_rd		; DR_RD
	call_oz(os_dor)
	ld a,dr_fre
	call_oz(os_dor)
	pop de			; restore IP
	ld ix,usr_zpad
	ex (sp),ix		; restore RP, put 2OS=x
	ld bc,0			; TOS=0, file exists
	jp EXIT


;X FLUSH-FILE		fileid -- ior
;   FILE-POSITION >R 2DROP R> ;		just checks fileid is valid
    head FLUSHFILE,10,"FLUSH-FILE",docolon,0,0
	defw FILEPOSITION,TOR,TWODROP,RFROM,EXIT


;X RENAME-FILE		caddr1 u1 caddr2 u2 -- ior
;   127 MIN STR>NUL DUP 128 + 128 CMOVE
;   127 MIN STR>NUL <<code>> ;
    head RENAMEFILE,11,"RENAME-FILE",docolon,0,0
	defw LIT,127,MIN,STRTONUL
	defw DUP,LIT,128,PLUS,LIT,128,CMOVE
	defw LIT,127,MIN,STRTONUL
	defw renam1
.renam1 push de                 ; save IP
	ld h,b
	ld l,c
	ld b,0			; BHL=address of original name
	ld de,usr_zpad+128	; DE=address of new name
	call_oz(gn_ren)
	pop de			; restore IP
	jp setior		; go to set success/error


; Dataspace words ================================

; In the Z88, we have the following dataspaces available:
;	RAM:	RAM, from 2000h to BFFFh
;	ROM2:	ROM in segment 2, from 8000h to BFFFh
;	ROM1:   ROM in segment 1, from 4000h to 7FFFh
; Note that ROM2 and ROM1 overlap with RAM, and are designed to allow
; applications that need more than a single 16K segment of ROM, not
; requiring a full 40K bad application space.


;Z RAM		--		; use RAM dataspace
;   DPS CURDS ! ;
    head RAM,3,"RAM",docode,0,ramn-ramf
.ramf   ld hl,usr_dps
	ld (usr_curds),hl
.ramn
	next

;Z ROM2		--		; use segment 2 ROM dataspace
;   DPS 2 + CURDS ! ;
    head ROM2,4,"ROM2",docode,0,rom2n-rom2f
.rom2f  ld hl,usr_dps+2
	ld (usr_curds),hl
.rom2n
	next

;Z ROM1		--		; use segment 1 ROM dataspace
;   DPS 4 + CURDS ! ;
    head ROM1,4,"ROM1",docode,0,rom1n-rom1f
.rom1f  ld hl,usr_dps+4
	ld (usr_curds),hl
.rom1n
	next

;X UNUSED	-- u		; unused space in current dataspace
;   CURDS @ DPS 4 + = IF  8000  ELSE  C000  THEN
;   HERE - ;
    head UNUSED,6,"UNUSED",docolon,0,0
	defw CURDS,FETCH,LIT,usr_dps+4,EQUAL
	defw QBRANCH,unuse1
	defw LIT,$8000,BRANCH,unuse2
.unuse1 defw LIT,$0C000
.unuse2 defw HERE,MINUS,EXIT


; Initial user variables =========================

;Z UINIT    -- addr  initial values for user area
;                    Unspecified values are ERASEd by COLD
    head UINIT,5,"UINIT",docode,0,uinitn-uinitf
.uinitf push    bc
        ld      bc,(uinitadd)
.uinitn
	next

; The standard CamelForth initial variables

.stduinit
	defw -1		; PENDING-CHAR
	defw CAMEL	; (COLD) vector
	defw BYE	; (RC_QUIT) vector
	defw CINIT	; (RC_DRAW) vector
	defw ESCAPE	; (RC_ESC) vector
	defw 0,0	; HOLE (no hole)
	defw DROP	; (KEY) vector
	defw DROP	; (ACC_EVT) vector
	defw NOOP	; (ACC_MAIL) vector
	defw lastword	; LATEST
	defw usr_wids	; NEXTLIST
	defw lastword	; FORTH-WORDLIST
	defw PKGMIA	; (RC_PNF) vector
	defw lastenv	; E0
	defw $2500	; DPS 0==RAM
	defw $8000	;     1==ROM2
	defw $4000	;     2==ROM1
	defw PKGERR	; (PKGERR) vector
	defw 0,0,0,0	; 8 new wordlists
	defw 0,0,0,0
	defw CEMIT	; EMIT
	defw CKEY	; KEY
	defw CTYPE	; TYPE
	defw NOOP	; PAUSE
	defw BRRUN	; RUN
	defw EKEYS	; EKEY
	defw BRPW	; PW
	defw 0		; 'BUFFER
	defw usr_forth	; CURRENT
	defw 1		; SORDER (9 cells)
	defw usr_forth
	defw 0,0,0,0,0,0,0
	defw usr_dps+0	; CURNS (RAM)
	defw usr_dps+0	; CURDS (RAM)
	defb 0		; #BUFS
	defw pertask_start	; LINK
	defb 0		; GBOUNDS
	defw 10		; BASE
	defw 0		; S0
	defw usr_rstacktop	; R0


;Z #INIT    -- n    #bytes of user area init data
    head NINIT,5,"#INIT",docon,0,0
        defw usr_initsize


; MAP GRAPHICS ROUTINES ==========================

;Z EXP?	-- flag				Is this Z88 expanded?
     head EXPQ,4,"EXP?",docode,0,notexp-expqf
.expqf  push ix
	push de
	ld ix,-1
	ld de,0
	ld a,3		; FA_EOF
	call_oz(os_frm)
	pop de
	pop ix
	push bc
	ld bc,0
	jr nz,notexp	; move on if not expanded
	dec bc
.notexp
	next

; Subroutine to find pixel address HL=x, BC=y
; Returns TOS=corrupt, HL=address, A=pixel, C=segment 2 binding
; and segment 2 bound to graphics area
; Sets carry if pixel out of range on map
; Enter at pixnob to avoid point being fixed to window bounds when
; GBOUND set

.pixadd ld a,(iy+usr_gbounds)
	and a
	jr z,pixnob	; move on if not forcing points in bounds
	ld a,h
	and a
	jr z,pixxok	; move on if X ok
	rlca
	ld hl,0		; fix to 0 if negative
	jr c,pixxok
	ld hl,255	; or 255 if >255
.pixxok ld a,b
	and a
	jr nz,pixybd	; move on if Y bad
	ld a,c
	cp 64
	jr c,pixnob	; move on if Y ok
.pixybd rlc b
	ld bc,0		; fix to 0 if negative
	jr c,pixnob
        ld bc,63        ; or 63 if >63
.pixnob ld (iy+usr_coords),c  ; set last point plotted (even if out of range)
	ld (iy+usr_coords+1),b
	ld (iy+usr_coords+2),l
	ld (iy+usr_coords+3),h
	ld a,b
	or h
	scf
	ret nz		; exit with carry set if X or Y >255
	ld b,l          ; B=x, C=y
	ld a,c
        cp 64
        ccf
        ret c           ; exit with carry set if Y out of range
        xor 63          ; convert to hardware arrangement (0 at top)
        ld c,a
        ld h,c
        srl h
        srl h
        srl h           ; linedist=y/8*256
        ld a,b
        and 248         ; rowdist=x/8*8
        ld l,a          ; HL=linedist+rowdist
        push hl
        ld a,c
        and 7           ; coldist=y MOD 8
        ld hl,(basegr)
        ld l,a          ; HL=graphicsarea+coldist
        ld a,b          ; A=x
        pop bc
        add hl,bc       ; HL=graphicsarea+coldist+linedist+rowdist
        and 7
        xor 7           ; A=7-(x MOD 8)
        ld b,a
        ld a,(seg2)     ; get current bank 2 binding to C
        ld c,a
        ld a,(basegr+2)
	and a
	scf
	ret z		; exit with carry set if no graphics bank
        ld (seg2),a     ; bind graphics to segment 2
        out (segp2),a
        ld a,1
	and a
	inc b
	dec b
        ret z
.pixad1 rlca
        djnz pixad1
        ret             ; exit with carry reset

;Z GBOUND	flag --		If true, make all graphic operations
;                               fall within window
    head GBOUND,6,"GBOUND",docode,0,gbndn-gbndf
.gbndf  ld a,b
	or c
	ld (iy+usr_gbounds),a
	pop bc
.gbndn
	next

;Z GSET         x y --          Set pixel at (x,y)
    head GSET,4,"GSET",docode,0,gsetn-gsetf
.gsetf  pop hl          ; HL=x, BC=y
        call pixadd
        jr c,pset1      ; no effect if out of range
        or (hl)
        ld (hl),a       ; plot point
        ld a,c
        ld (seg2),a     ; rebind segment 2
        out (segp2),a
.pset1  pop bc          ; get new TOS
.gsetn
	next


;Z GCLR         x y --          Clear pixel at (x,y)
    head GCLR,4,"GCLR",docode,0,gclrn-gclrf
.gclrf  pop hl          ; HL=x, BC=y
        call pixadd
        jr c,pclr1      ; no effect if out of range
        cpl
        and (hl)
        ld (hl),a       ; reset point
        ld a,c
        ld (seg2),a     ; rebind segment 2
        out (segp2),a
.pclr1  pop bc          ; get new TOS
.gclrn
	next


;Z GINV         x y --          Invert pixel at (x,y)
    head GINV,4,"GINV",docode,0,ginvn-ginvf
.ginvf  pop hl          ; HL=x, BC=y
        call pixadd
        jr c,pinv1      ; no effect if out of range
        xor (hl)
        ld (hl),a       ; invert point
        ld a,c
        ld (seg2),a     ; rebind segment 2
        out (segp2),a
.pinv1  pop bc          ; get new TOS
.ginvn
	next

;Z GSET?	x y --	flag	Test if pixel at (x,y) set
    head GSETQ,5,"GSET?",docode,0,gsetqn-gsetqf
.gsetqf pop hl          ; HL=x, BC=y
	call pixadd
	ld b,0		; false flag
	jr c,gstq2	; use false if out of range
	and (hl)
	jr z,gstq1
	dec b
.gstq1  ld a,c
	ld (seg2),a	; rebind segment 2
	out (segp2),a
.gstq2  ld c,b          ; make flag
.gsetqn
	next

;Z GMOVE	x y --		Move plot position to (x,y)
    head GMOVE,5,"GMOVE",docode,0,gmoven-gmovef
.gmovef pop hl
	ld (iy+usr_coords),c
	ld (iy+usr_coords+1),b
	ld (iy+usr_coords+2),l
	ld (iy+usr_coords+3),h
	pop bc		; new TOS
.gmoven
	next

;Z GPOS?	-- x y		Get plot position
    head GPOSQ,5,"GPOS?",docode,0,gposqn-gposqf
.gposqf push bc         ; save old TOS
	ld l,(iy+usr_coords+2)
	ld h,(iy+usr_coords+3)
	push hl
	ld c,(iy+usr_coords)
	ld b,(iy+usr_coords+1)
.gposqn
	next

;Z (OPENMAP)      id --       Open window to 256-pixel map
    nohead BROPENMAP,docode
	ld hl,255	; width of window to use
        ld a,c
        add a,'0' ; A=window number
        ld bc,3		; MP_GRA
        call_oz(os_map)
	jr c,opmap2	; move on if couldn't
        ld hl,0
        ld b,h
        ld a,sc_hr0
        call_oz(os_sci)
        push bc
        push hl
        call_oz(os_sci)
        pop hl
        pop bc
        set 7,h
        res 6,h         ; force segment 2 address
        ld (basegr),hl  ; save address
        ld a,b
        ld (basegr+2),a ; save bank
.opmap2 pop bc          ; get new TOS
	next

;Z OPENMAP	id --		Open 256-pixel map
;   EXP? IF  (OPENMAP) GPAGE  ELSE  DROP  THEN ;
    head OPENMAP,7,"OPENMAP",docolon,0,0
	defw EXPQ,QBRANCH,opmp1
	defw BROPENMAP,GPAGE,EXIT
.opmp1  defw DROP
.opmp2  defw EXIT

;Z GPAGE	--		Clear graphics area
    head GPAGE,5,"GPAGE",docode,0,gpagen-gpagef
.gpagef ld a,(seg2)
	push af		; save segment 2 binding
	ld a,(basegr+2)
	and a
	jr z,gpage2	; move on if no graphics area
	ld (seg2),a
	out (segp2),a	; page in graphics area
	push bc
	push de
	ld hl,(basegr)
	ld (hl),0
	ld d,h
	ld e,l
	inc de
	ld bc,256*64/8-1	; size of graphics area in bytes
	ldir		; clear it
	pop de
	pop bc
.gpage2 pop af
	ld (seg2),a	; repage original
	out (segp2),a
	xor a
	ld (iy+usr_coords),a	; reset last plotted coords
	ld (iy+usr_coords+1),a
	ld (iy+usr_coords+2),a
	ld (iy+usr_coords+3),a
.gpagen
	next


; Subroutine to draw a line from last coordinates across HL=X, up BC=Y

.drawline ld a,b
	rlca
	ld a,c
	ld d,1
	jr nc,gline1	; move on if Y positive
	neg
	ld d,-1
.gline1 ld b,a          ; B=ABS Y, D=SGN Y
	ld a,c
	and a
	jr nz,glin1a
	ld d,0		; if ABS Y=0, SGN Y=0
.glin1a ld a,h
	rlca
	ld a,l
	ld e,1
	jr nc,gline2	; move on if X positive
	neg
	ld e,-1
.gline2 ld c,a          ; C=ABS X, E=SGN X
	ld a,l
	and a
	jr nz,glin2a
	ld e,0		; if ABS X=0, SGN X=0
.glin2a ld a,c
	cp b
	jr nc,gline3	; find smaller of ABS X and ABS Y
	ld l,c		; ABS X is smaller
	push de		; save diagonal step
	ld e,0		; use vertical step
	jr gline4
.gline3 or b
	ret z		; exit if nothing to do
	ld l,b		; ABS Y is smaller
	ld b,c		; ABS X is larger 
	push de		; save diagonal step
	ld d,0		; use horizontal step
.gline4 ld h,b          ; largest of ABS X/ABS Y
	ld a,b
	rra
.gline5 add a,l
	jr c,gline6
	cp h
	jr c,gline7
.gline6 sub h
	ld c,a
	exx
	pop bc		; BC'=diagonal step
	push bc
	jr gline8
.gline7 ld c,a
	push de
	exx
	pop bc		; BC'=hor/vert step
.gline8 ld l,(iy+usr_coords)  ; HL=last Y
	ld h,(iy+usr_coords+1)
	ld d,b
	dec d
	jr z,glinec
	inc d
.glinec ld e,b          ; DE=vertical step (0, -1 or 1)
	add hl,de
	ld d,c
	dec d
	jr z,glined
	inc d
.glined ld e,c          ; DE=horizontal step (0, -1 or 1)
	ld b,h
	ld c,l		; BC=new Y
	ld l,(iy+usr_coords+2) ; HL=last X
	ld h,(iy+usr_coords+3)
	add hl,de	; HL=new X
	call pixadd
	jr c,glinea	; don't plot if out of range
        or (hl)
        ld (hl),a       ; plot point
.glinea ld a,c
	ld (seg2),a     ; rebind segment 2
        out (segp2),a
	exx
	ld a,c
	djnz gline5
	pop de		; discard diagonal step
	ret

;Z GLINE	x y --		Draw line from last plot across X, up Y
    head GLINE,5,"GLINE",docode,0,glinen-glinef
.glinef pop hl          ; HL=X, BC=Y
	push de		; save IP
	call drawline
	pop de		; restore IP
	pop bc		; new TOS
.glinen
	next

;Z GLINETO	x y --		Draw line from last pos TO (x,y)
;   GPOS?
;   NEGATE ROT + SWAP
;   NEGATE ROT + SWAP
;   GLINE ;
    head GLINETO,7,"GLINETO",docolon,0,0
	defw GPOSQ
	defw NEGATE,ROT,PLUS,SWOP
	defw NEGATE,ROT,PLUS,SWOP
	defw GLINE,EXIT

;Z GBOX		w h --		Draw box
;   OVER 0 GLINE  0 OVER GLINE
;   SWAP NEGATE 0 GLINE
;   NEGATE 0 SWAP GLINE ;	
    head GBOX,4,"GBOX",docolon,0,0
	defw OVER,ZERO,GLINE,ZERO,OVER,GLINE
	defw SWOP,NEGATE,ZERO,GLINE
	defw NEGATE,ZERO,SWOP,GLINE,EXIT

;Z GELLIPSE	x y w h --	Draw ellipse
;   SWAP ABS SWAP ABS 1 MAX
;   DUP 8 * 360 MIN		; number of steps
;   360 SWAP / >R
;   DUP >R 2OVER R> + GSET R>	; set point at top
;   360 0 DO  usr_scratch !
;             2OVER 2OVER 
;	      I COS* ROT + SWAP
;             I SIN* ROT + SWAP
;             GLINETO
;             usr_scratch @ PAUSE
;         DUP +LOOP
;   DROP NIP + GLINETO DROP ;	; finish at top
    head GELLIPSE,8,"GELLIPSE",docolon,0,0
	defw SWOP,ABS,SWOP,ABS,ONE,MAX
	defw DUP,LIT,8,STAR,LIT,360,MIN
	defw LIT,360,SWOP,SLASH,TOR
	defw DUP,TOR,TWOOVER,RFROM,PLUS,GSET,RFROM
	defw LIT,360,ZERO,XDO
.gellp1 defw LIT,usr_scratch,STORE
	defw TWOOVER,TWOOVER
	defw II,COSSTAR,ROT,PLUS,SWOP
	defw II,SINSTAR,ROT,PLUS,SWOP
	defw GLINETO
	defw LIT,usr_scratch,FETCH,PAUSE
	defw DUP,XPLUSLOOP,gellp1
	defw DROP,NIP,PLUS,GLINETO,EXIT

;Z GCIRCLE	x y r --	Draw circle
;   DUP GELLIPSE ;
    head GCIRCLE,7,"GCIRCLE",docolon,0,0
	defw DUP,GELLIPSE,EXIT


; Subroutine to test if pixel at HL=x, BC=y is set
; Preserves BC & HL. NZ if pixel set.

.pixtest push bc
	push hl
	call pixnob
	jr c,pixt2
	and (hl)
	ld a,c
	ld (seg2),a
	out (segp2),a
.pixt2  pop hl
	pop bc
	ret

; Subroutine to fill a horizontal line with pixels
; HL=x, BC=y (preserved)

.linefill push ix       ; save shade address
	ld d,0
	ld a,c
	and 7
	ld e,a
	add ix,de	; point to correct shade address
	ld a,(ix+0)
	ex af,af'	; A'=shade (must save as it may be in segment 2)
	push hl	; save initial X
.lfill1 dec hl          ; left
	ld a,h
	and a
	jr nz,lfill2	; move on if out of range
	push hl
	push bc		; save coords
	call pixnob
	jr c,lfill3
	push af
	and (hl)
	jr nz,lfill4	; move on if pixel already set
	ex af,af'
	ld b,a		; B=shade
	ex af,af'
	pop af
	and b		; mask with shade
	or (hl)
	ld (hl),a	; set pixel
	ld a,c
	ld (seg2),a
	out (segp2),a
.lfill3 pop bc
	pop hl
	jr lfill1	; back for more leftward pixels
.lfill4 pop af          ; found bounding pixel, so tidy up
	ld a,c
	ld (seg2),a
	out (segp2),a
	pop bc
	pop hl
.lfill2 pop hl          ; restore original X
	push hl		; and re-save
.lfill5 ld a,h
	and a
	jr nz,lfill6	; move on if out of range
	push hl
	push bc		; save coords
	call pixnob
	jr c,lfill7
	push af
	and (hl)
	jr nz,lfill8	; move on if pixel already set
	ex af,af'
	ld b,a		; B=shade
	ex af,af'
	pop af
	and b		; mask with shade
	or (hl)
	ld (hl),a	; set pixel
	ld a,c
	ld (seg2),a
	out (segp2),a
.lfill7 pop bc
	pop hl
	inc hl		; right
	jr lfill5	; back for more rightward pixels
.lfill8 pop af          ; found bounding pixel, so tidy up
	ld a,c
	ld (seg2),a
	out (segp2),a
	pop bc
	pop hl
.lfill6 ld ix,(usr_scratch)	; get RP
	ld de,lfill9
	jp PAUSE	; allow other tasks to run
.lfill9	defw lfill9+2
	ld (usr_scratch),ix	; save RP again
	pop hl          ; restore original X
	pop ix		; restore shade address
	ret

;Z GPATTERN	x y c-addr --		Fill area with pattern
    head GPATTERN,8,"GPATTERN",docolon,0,0
	defw gpatst
.gpatst	pop hl
	ex de,hl
	ex (sp),hl	; DE=y, HL=x, IP saved
	ld (usr_scratch),ix	; save RP
	push bc
	pop ix		; IX=shade address
	ld b,d
	ld c,e		; BC=y
	push hl		; save coords
	push bc
.gfill1 inc bc          ; up
	ld a,c
	and 192
	or b
	jr nz,gfill2	; move on if Y out of range
	call pixtest
	jr nz,gfill2	; or if pixel set
	call linefill
	jr gfill1
.gfill2 pop bc
	pop hl		; get original coords
	push hl		; and resave
	push bc
.gfill3 ld a,c
	and 192
	or b
	jr nz,gfill4	; move on if Y out of range
	call pixtest
	jr nz,gfill4	; or if pixel set
	call linefill
	dec bc		; down
	jr gfill3
.gfill4 pop bc
	pop hl		; get original coords
	ld d,0
	ld a,c
	and 7
	ld e,a
	add ix,de	; get correct shade
	call pixnob	; get pixel address (no bounds)
	jr c,gfill5
	and (ix+0)	; mask with shade
	or (hl)
	ld (hl),a	; set it
	ld a,c
	ld (seg2),a
	out (segp2),a
.gfill5 ld ix,(usr_scratch)	; restore RP
	pop de		; restore IP
	pop bc		; new TOS
	jp EXIT

;Z GFILL	x y --		Fill area with solid pixels
;   shdblk GPATTERN ;
    head GFILL,5,"GFILL",docolon,0,0
	defw LIT,shdblk,GPATTERN,EXIT

.shdblk defb 255,255,255,255,255,255,255,255

;Z GSHADE	x y --		Shade area with alternate pixels
;   shdgrey GPATTERN ;
    head GSHADE,6,"GSHADE",docolon,0,0
	defw LIT,shdgrey,GPATTERN,EXIT

.shdgrey defb 170,85,170,85,170,85,170,85


; TEXT WINDOW ROUTINES ===========================

; Subroutine to open a text window with: x y w h ID+256*type
; IX=offset into stack for height

.opwin  ld hl,windef
	call_oz(gn_sop)
	ld a,c
     	add a,'0'
	call_oz(os_out)
     	ld a,(ix+6)
	add a,32
	call_oz(os_out)
     	ld a,(ix+4)
	add a,32
	call_oz(os_out)
     	ld a,(ix+2)
	add a,32
	call_oz(os_out)
     	ld a,(ix+0)
	add a,32
	call_oz(os_out)
     	ld a,b
	call_oz(os_out)
	ld hl,winclr
	call_oz(gn_sop)
	ld a,c
	add a,'0'
	call_oz(os_out)
	ld hl,winmod	
	call_oz(gn_sop)
	ret

.windef defm 1&"7#"&0     ; window definer
.winclr defm 1&"2C"&0     ; window clear & reset modes
.winmod defm 1&"3+CS"&0   ; set default modes (cursor & scroll)

;Z OPENWINDOW  x y w h id --        Open normal window
    head OPENWINDOW,10,"OPENWINDOW",docode,0,opwinn-opwinf
.opwinf push ix
	ld ix,2
	add ix,sp
	ld b,129	; left and right bars only
	call opwin	; open the window
	pop ix
.opdisc pop bc          ; discard x,y,w,h
	pop bc
	pop bc
	pop bc	
     	pop bc         ; get new TOS
.opwinn
	next

;Z OPENTITLED	x y w h c-addr id --	Open titled window
    head OPENTITLED,10,"OPENTITLED",docode,0,0
	push ix
	ld ix,4
	add ix,sp
.wintitle ld b,131      ; with shelf brackets
	call opwin	; open the window
	ld hl,winban
	call_oz(gn_sop)
	ld l,(ix-2)
	ld h,(ix-1)	; address of banner
	call_oz(gn_sop)
	ld hl,winapl	; apply effects string
	call_oz(gn_sop)
	ld a,(ix+2)
	add a,'0'
	call_oz(os_out)
	inc (ix+4)	; increment row start
	dec (ix+0)	; decrement window height
	ld b,129
	call opwin	; re-open with left & right bars only
	pop ix
	pop bc		; discard banner address
	jr opdisc	; discard other parameters and exit	

.winbot defm 1&"2-S"&1&"2Y"&0
.winlin defm 1&"2*"&73&1&"3N"&0
.wincrn defm 5&1&"2*"&76&0
.winban defm 1&"4+TUR"&1&"2JC"&0
.winapl defm 1&"3@  "&1&"2A"&0

;Z OPENPOPUP	x y w h c-addr id --	Open popup window
    head OPENPOPUP,9,"OPENPOPUP",docode,0,0
	push ix
	ld ix,4
	add ix,sp
	dec (ix+6)
	inc (ix+2)
	inc (ix+2)
	ld b,128	; with nothing
	call opwin	; open the window
	inc (ix+6)
	dec (ix+2)
	dec (ix+2)
	ld hl,winbot	; get to bottom of window
	call_oz(gn_sop)
	ld a,(ix+0)
	add a,31
	call_oz(os_out)
	ld hl,winlin	; start of bottom line
	call_oz(gn_sop)
	ld a,(ix+2)
	add a,32
	call_oz(os_out)
	ld hl,wincrn	; corner of window
	call_oz(gn_sop)
	dec (ix+0)	; reduce height of window
        jp wintitle     ; carry on with OPENTITLED

;Z WINDOW	id --		Select window for output
    head WINDOW,6,"WINDOW",docode,0,windwn-windwf
.windwf ld hl,winsel
	call_oz(gn_sop)
	ld a,c
	add a,'0'
	call_oz(os_out)
	pop bc		; get new TOS
.windwn
	next

.winsel defm 1&"2H"&0

;Z WINDOW?	-- w h id	Enquire current window parameters
    head WINDOWQ,7,"WINDOW?",docode,0,windqn-windqf
.windqf push bc         ; stack old TOS
	push de		; save IP
	ld bc,nq_wbox
	xor a
	call_oz(os_nq)
	pop de		; restore IP
	ld l,c
	ld h,0
	push hl		; stack width
	ld l,b
	push hl		; stack height
	ld b,h
	sub '0'
	ld c,a		; TOS=id
.windqn
	next

;Z EFFECT	char1 char2 --		Set text effect
;   1 EMIT [CHAR] 2 EMIT EMIT EMIT ;
    nohead EFFECT,docolon
	defw ONE,EMIT,LIT,50,EMIT,EMIT,EMIT,EXIT

;Z ESET	char --		Set text effect (eg use CHAR U ESET to set underlining)
;   [CHAR] + EFFECT ;
    head ESET,4,"ESET",docolon,0,0
     defw LIT,43,EFFECT,EXIT

;Z ECLR	char --		Clear text effect
;   [CHAR] - EFFECT ;
    head ECLR,4,"ECLR",docolon,0,0
     defw LIT,45,EFFECT,EXIT

;Z ETOG	char --		Toggle text effect
;   1 EMIT EMIT ;
    head ETOG,4,"ETOG",docolon,0,0
	defw ONE,EMIT,EMIT,EXIT

;Z EAPPLY	n --	Apply text effects
;   BL + [CHAR] A EFFECT ;
    head EAPPLY,6,"EAPPLY",docolon,0,0
	defw BL,PLUS,LIT,65,EFFECT,EXIT

;Z EAPPLYX	n --	Apply text effects with XOR
;   BL + [CHAR] E EFFECT ;
    head EAPPLYX,7,"EAPPLYX",docolon,0,0
	defw BL,PLUS,LIT,69,EFFECT,EXIT

;Z EALIGN	char --		Set alignment
;   [CHAR] J EFFECT ;
    head EALIGN,6,"EALIGN",docolon,0,0
	defw LIT,74,EFFECT,EXIT

;Z GREY		--		Grey current window
;   ." ..." ;
    head GREY,4,"GREY",docolon,0,0
	defw XSQUOTE
	defb 4
	defm 1&"2G+"
	defw TYPE,EXIT

;Z UNGREY	--		Ungrey current window
;   ." ..." ;
    head UNGREY,6,"UNGREY",docolon,0,0
	defw XSQUOTE
	defb 4
	defm 1&"2G-"
	defw TYPE,EXIT  

;Z SCRUP		--		Scroll window up
;   ." ..." ;
    head SCRUP,5,"SCRUP",docolon,0,0
	defw XSQUOTE
	defb 2
	defb 1,255
	defw TYPE,EXIT

;Z SCRDOWN	--		Scroll window down
;   ." ..." ;
    head SCRDOWN,7,"SCRDOWN",docolon,0,0
	defw XSQUOTE
	defb 2
	defb 1,254
	defw TYPE,EXIT

;Z DEFUDG	addr char --		Define a UDG
    head DEFUDG,6,"DEFUDG",docode,0,defudn-defudf
.defudf ld hl,defu
	call_oz(gn_sop)
	ld a,c
	call_oz(os_out)
	pop hl
	ld b,8
.defud1 ld a,(hl)
	and 63
	or 128
	inc hl
	call_oz(os_out)
	djnz defud1
	pop bc		; get new TOS
.defudn
	next

.defu   defm 1&138&"="&0

;Z UDG	char --		Output UDG
;   [CHAR] ? EFFECT ;
    head UDG,3,"UDG",docolon,0,0
	defw LIT,63,EFFECT,EXIT

;Z XCHAR char --	Output special character
;   ETOG ;
    head XCHAR,5,"XCHAR",docolon,0,0
	defw ETOG,EXIT

;Z XLINE char --	Output line character (A-O)
;   [CHAR] * EFFECT ;
    head XLINE,5,"XLINE",docolon,0,0
	defw LIT,42,EFFECT,EXIT

;Z CINIT --         initialise the console window
;   1 0 92 8  1 OPENWINDOW ;
    head CINIT,5,"CINIT",docolon,0,0
	defw ONE,ZERO,LIT,92,LIT,8,ONE,OPENWINDOW,EXIT

;Z PW   --                      perform a page wait (deferred)
;   DEFER PW
    head PW,2,"PW",dodefer,0,0
	defw usr_pw

;Z (PW)	--			perform a page wait
    head BRPW,4,"(PW)",docode,0,0
	ld a,sr_pwt
	call_oz(os_sr)
	call vector	; error handling
	next

; Random numbers =================================

;Z RAND	u --		set the random number seed
    head RAND,4,"RAND",docode,0,randmn-randmf
.randmf ld a,b
        or c
	jr nz,rand1	; use seed if provided
	ld c,$0D0
	in a,(c)
	inc c
	in b,(c)	; else use RTC
.rand1  ld (usr_seed),bc
	pop bc		; new TOS
.randmn
	next

;Z RND	u1 -- u2	get random number between 0 and u1-1
;   20077 16838 usr_seed @ UD*
;   12345 0 D+
;   OVER usr_seed !
;   ROT UD/MOD 2DROP ;
    head RND,3,"RND",docolon,0,0
	defw LIT,20077,LIT,16838
	defw LIT,usr_seed,FETCH,UDSTAR
	defw LIT,12345,ZERO,DPLUS
	defw OVER,LIT,usr_seed,STORE
	defw ROT,UDSLASHMOD,TWODROP,EXIT


; New definition of WORDS, using PW ========
; We now define this in terms of FWORDS, which lists words within
; a defined range of "fast" code lengths. All fast words can be listed with
; 128 1 FWORDS, since the largest fast length is 127 bytes, and all words
; can be listed with 128 0 FWORDS, since non-fast words have a length of zero.
; In general, use h l FWORDS, where l is the shortest length and h is the
; largest length+1.

;Z FWORDS  u1 u2  --          list fast words between lengths u1 & u2 in dict.
;   SWAP 2>R 0 SORDER @
;   IF  CR WINDOW? DROP *
;	SORDER CELL+ @
;       BEGIN  @ ?DUP
;       WHILE  DUP COUNT + C@ 127 AND 2R@ WITHIN
;              IF
;                DUP C@ 1+ ROT SWAP -
;	         TUCK 1- 0< IF  NIP WINDOW? DROP * OVER C@ 1+ - SWAP
;                               PW CR  
;                           THEN
;                DUP COUNT TYPE SPACE
;		 ROT 1+ ROT ROT
;              THEN
;              2 -
;       REPEAT
;       WINDOW? 2DROP 2* < IF  PW  THEN
;   THEN  2R> 2DROP
;   CR . ." words" ;
    head FWORDS,6,"FWORDS",docolon,0,0
	defw SWOP,TWOTOR,ZERO,SORDER,FETCH,QBRANCH,WDS2
	defw CRR,WINDOWQ,DROP,STAR
	defw SORDER,CELLPLUS,FETCH
.WDS1   defw FETCH,QDUP
	defw QBRANCH,WDS4
	defw DUP,COUNT,PLUS,CFETCH,LIT,127,AND,TWORFETCH,WITHIN
	defw QBRANCH,WDS5
	defw DUP,CFETCH,ONEPLUS,ROT,SWOP,MINUS
	defw TUCK,ONEMINUS,ZEROLESS,QBRANCH,WDS3
	defw NIP,WINDOWQ,DROP,STAR,OVER,CFETCH,ONEPLUS,MINUS,SWOP
	defw PW,CRR
.WDS3   defw DUP,COUNT,TYPE,SPACE
	defw ROT,ONEPLUS,ROT,ROT
.WDS5	defw LIT,2,MINUS
	defw BRANCH,WDS1
.WDS4	defw WINDOWQ,TWODROP,TWOSTAR,LESS,QBRANCH,WDS2
	defw PW
.WDS2   defw TWORFROM,TWODROP,CRR,DOT,XSQUOTE
	defm 5&"words"
	defw TYPE,EXIT

;X WORDS	--	      list words
;    128 0 FWORDS ;
     head WORDS,5,"WORDS",docolon,0,0
         defw LIT,128,ZERO,FWORDS,EXIT


; New far memory words ======================
; Far memory is now accessed via double-cell addresses (high byte
; is ignored, so effectively a 3-byte address) in the same scheme as
; implemented for Small C on the Z88. Local addresses are accessed when
; the high byte is zero.
;
; Before accessing any far memory, a pool must be created using POOL
; and initialised by executing the pool word. ONLY ONE pool is permitted;
; attempting to use more than one will result in resource leakage and
; crashes... 

defc    seg1    =     $4D1
defc    segp1   =     $0D1

; POOL	ud "name" --	        create pool of up to ud bytes
;                                   Executing the word created initialises it
;   NSDS CREATE  255 AND
;	    256 UM/MOD
;	    SWAP IF  1+  THEN
;	    DUP , HERE 0 ,
;           CURDS @ RAM HERE ROT !
;	    SWAP 2* ALLOT
;           CURDS !
;   DOES> ( -- )  DUP @ farpages ! CELL+ @ malloc_table ! 
;		  $40 farmemspec C!
;		  pool_table 224 ERASE
;		  malloc_table @ farpages @ 2* ERASE ;
    head POOL,4,"POOL",docolon,0,0
	defw NSDS,CREATE,LIT,255,AND
	defw LIT,256,UMSLASHMOD
	defw SWOP,QBRANCH,mempl1,ONEPLUS
.mempl1 defw DUP,COMMA,HERE,ZERO,COMMA
	defw CURDS,FETCH,RAM,HERE,ROT,STORE
	defw SWOP,TWOSTAR,ALLOT
	defw CURDS,STORE,XDOES
.domem  call dodoes
	defw DUP,FETCH,LIT,farpages,STORE
	defw CELLPLUS,FETCH,LIT,malloc_table,STORE
	defw LIT,$40,LIT,farmemspec,CSTORE
	defw LIT,pool_table,LIT,224,ERASE
	defw LIT,malloc_table,FETCH,LIT,farpages,FETCH,TWOSTAR,ERASE
	defw EXIT

; Subroutine to bind far address CDE to segment 1 at address HL.
; On entry, A' contains binding of segment 1 for local memory (preserved).

.faradd ex      af,af'
	ld	(seg1),a
	out	(segp1),a	; ensure local memory bound to segment 1
	ex	af,af'
	ld	a,c
	and	a
	jr	z,locfar	; move on for local pointers
	ld	hl,(malloc_table)
	dec	c
	ld	b,c
	ld	c,d
	add	hl,bc
	add	hl,bc		; HL points to 2-byte entry
	ld	c,b
	inc	c		; restore CDE
	ld	a,(hl)		; A=bank
	inc	hl
	ld	h,(hl)		; H=address high byte (in seg 1)
	ld	l,e		; low byte is the same
	ld	(seg1),a
	out	(segp1),a	; bind to segment 1
	ret
.locfar ld      h,d             ; address is just DE
	ld	l,e
	ret

; Subroutine to increment local address HL with far address CDE
; Corrupts B (via faradd) but preserves A.

.incfar inc     hl
	inc	e
	ret	nz
	inc	d
	jr	nz,skiphi
	inc	c
.skiphi push    af
	call	faradd
	pop	af
	ret

;Z ALLOCATE	ud -- dl-addr ior	Allocate extended memory
;  As the ANSI word ALLOCATE, except size and address are both double-cells
     head ALLOCATE,8,"ALLOCATE",docode,0,0
	ex	de,hl
	ex	(sp),hl		; save IP; BCHL=required size (bytes)
	push	ix		; save RP
	ld	a,c
	and	128
	or	b
	jp	nz,badmalloc	; trying to allocate 8Mb+....
	ld	a,l
	ld	b,c
	ld	c,h		; BC=pages required, ignoring low byte
	inc	bc
	inc	a
	jr	nz,mall2
	inc	bc		; BC=total pages (with 2-byte overhead)

; Stage 1 is to find a free fragment in the malloc table

.mall2  ld      hl,(malloc_table)
	ld	ix,0		; size found
	ld	de,(farpages)
	push	ix		; save best match so far
	push	hl
.malsearch
	ld	a,d
	or	e
	jr	z,endsearch	; end of malloc_table reached
	dec	de
	ld	a,(hl)
	inc	hl
	inc	hl
	and	a
	jr	nz,malsearch	; back until found a free page
	push	hl		; save start of this fragment+2
	ld	ix,0
.malfragment
	inc	ix
	ld	a,d
	or	e
	jr	z,endfragment	; end of malloc_table reached
	dec	de
	ld	a,(hl)
	inc	hl
	inc	hl
	and	a
	jr	z,malfragment	; back until end of fragment
.endfragment
	push	ix
	ex	(sp),hl
	and	a
	sbc	hl,bc		; is fragment big enough?
	jr	nc,goodmatch
	pop	hl
	pop	af		; discard fragment start
	jr	malsearch
.goodmatch
	add	hl,bc
	ex	(sp),hl
	exx
	pop	hl		; HL'=fragment size
	pop	ix		; IX=fragment start+2
	dec	ix
	dec	ix		; IX=real fragment start
	pop	de		; DE'=start of best so far
	pop	bc		; BC'=size of best so far
	ld	a,b
	or	c
	jr	z,usenew	; use new fragment if no previous one
	sbc	hl,bc
	jr	nc,useold	; use old if <=size of new one
	add	hl,bc		; restore HL=fragment size
.usenew push    hl
	push	ix
	exx
	jr	malsearch
.useold push    bc
	push	de
	exx
	jr	malsearch
.endsearch
	pop	hl		; HL=start of best fragment found
	pop	de		; DE=size
	ld	a,d
	or	e
	jr	z,badmalloc	; no suitable fragment found

; Stage 2 is to fill the fragment with memory
; We start with HL=address in malloc_table & BC=pages required

	push	hl		; save parameters
	push	bc
	ld	d,31		; pretend bank 31 is available
.nextpool
	call	findpool	; IX=pool to allocate from, D=bank
	jr	c,partmalloc
.nextpage
	push	bc		; save pages to do
	push	hl		; save malloc_table address
.badpage
	xor	a
	ld	bc,256
	call_oz(os_mal)
	jr	nc,gotpage
	pop	hl
	pop	bc
	jr	nextpool
.gotpage
	ld	a,l
	and	a
	jr	nz,badpage	; EEK! OZ returned non-page-aligned mem...
	ld	a,b
	cp	d
	jr	nz,badpage	; EEK! Not from the expected bank...
	ld	a,h
	pop	hl
	ld	(hl),d		; store bank
	inc	hl
	ld	(hl),a		; and high byte
	inc	hl
	pop	bc
	dec	bc
	ld	a,b
	or	c
	jr	nz,nextpage
	pop	bc		; pages allocated
	pop	hl		; address in malloc_table
	ld	a,(seg1)
	ex	af,af'		; save seg1 binding
	ld	a,(hl)
	inc	hl
	ld	d,(hl)
	dec	hl
	ld	(seg1),a
	out	(segp1),a	; bind in start of allocated memory
	ld	e,0		; DE=address in segment 1
	ld	a,c
	ld	(de),a
	inc	de
	ld	a,b
	ld	(de),a		; store # pages allocated in 1st two bytes
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; rebind seg 1
	ld	de,(malloc_table)
	and	a
	sbc	hl,de
	srl	h
	rr	l		; HL=pool page number
	ld	b,0
	ld	c,h
	inc	c		; because C=0 is local memory
	ld	h,l
	ld	l,2		; BCHL=extended memory pointer
	pop	ix		; restore RP
	pop	de		; restore IP
	push	hl
	push	bc		; stack extended memory pointer
	ld	bc,0		; success!
	next

; At this point we have a partial allocation, so need to free up what
; we've allocated so far

.partmalloc
	pop	bc
	pop	hl
	call	free_loop

; Here we return ior=-263 (equivalent to OZ's RC_ROOM) on allocation
; failure

.badmalloc
	pop	ix		; restore RP
	pop	de		; restore IP
	ld	hl,0
	push	hl
	push	hl		; use dl-addr=0
	ld	bc,-263
	next

; Subroutine of ALLOCATE which finds a new pool to allocate from.
; IN:	D=previous bank
; OUT:	D=new bank
;	IX=pool handle
;	Fc=0 if successful, 1 if failed

.findpool
	push	bc
	push	hl
	ld	hl,pool_table-32
	ld	c,d
	ld	b,0
	add	hl,bc
.fpool2 inc     hl
	inc	d
	jr	z,openpool	; open new pool if end of bank list
	ld	a,(hl)
	and	a
	jr	z,fpool2	; back if pool not open yet
	add	a,a
	rl	b		; (B was 0 from code above)
	add	a,a
	rl	b
	add	a,a
	rl	b
	add	a,a
	rl	b
	ld	ixh,b		; ld ixh,b
	ld	ixl,a		; ld ixl,a - IX=pool handle
	and	a		; Fc=0, success!
.nopool pop     hl
	pop	bc
	ret
.openpool
	ld	a,(farmemspec)
	ld	bc,0
	call_oz(os_mop)
	jr	c,nopool
	xor	a
	ld	bc,256
	call_oz(os_mal)
	ld	a,b
	push	af		; save A=bank number
	ld	bc,256
	call_oz(os_mfr)
	ld	hl,pool_table-32
	pop	af
	ld	c,a
	ld	b,0		; BC=bank number
	add	hl,bc
	ld	a,ixl		; ld a,ixl
	ld	b,ixh		; ld b,ixh
	srl	b
	rra
	srl	b
	rra
	srl	b
	rra
	srl	b
	rra
	ld	(hl),a		; save compressed pool handle
	ld	d,c		; D=bank
	pop	hl
	pop	bc
	and	a		; success!
	ret

;Z FREE		dl-addr -- ior		Free extended memory
;  As the ANSI word FREE, except address is a double-cell
     head FREE,4,"FREE",docode,0,0
	ex	de,hl
	ex	(sp),hl		; save IP, BCHL=far pointer
	push	ix		; save RP
	ld	a,l
	cp	2
	jr	nz,badfree	; low byte=2 always
	dec	c		; because 0=local pointer
	ld	b,c
	ld	c,h		; BC=page number
	ld	hl,(malloc_table)
	add	hl,bc
	add	hl,bc		; HL=address in malloc_table
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
	ld	a,(hl)
	inc	hl
	ld	d,(hl)
	dec	hl
	ld	(seg1),a
	out	(segp1),a
	ld	e,0
	ld	a,(de)
	ld	c,a
	inc	de
	ld	a,(de)
	ld	b,a		; BC=#pages to deallocate
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; rebind seg 1
	call	free_loop	; free the memory
	ld	bc,0		; success!
.endfree
	pop	ix		; restore RP
	pop	de		; restore IP
	next
.badfree
	ld	bc,-9		; invalid memory address
	jr	endfree

; Subroutine used by ALLOCATE and FREE to free BC pages at a malloc_table
; address of HL

.free_loop
	ld	a,(hl)
	and	a
	jr	z,notallocated
	ld	e,a
	ld	d,0		; DE=bank
	ld	(hl),d
	inc	hl
	ld	ix,pool_table-32
	add	ix,de
	ld	a,(ix+0)	; A=compressed pool handle
	add	a,a
	rl	d		; D was 0 previously
	add	a,a
	rl	d
	add	a,a
	rl	d
	add	a,a
	rl	d
	ld	ixh,d		; ld ixh,d
	ld	ixl,a		; ld ixl,a - IX=pool handle
	ld	a,e		; A=bank
	ld	d,(hl)
	ld	e,0		; DE=address
	ld	(hl),e
	inc	hl
	push	bc
	ld	bc,256
	ex	de,hl
	call_oz(os_mfr)
	ex	de,hl
	pop	bc
.freenext
	dec	bc
	ld	a,b
	or	c
	jr	nz,free_loop
	ret
.notallocated
	inc	hl
	inc	hl
	jr	freenext

;Z FREEALL	--		Free all extended memory & handles
;  This word must be called if extended memory is used before exiting CF
     head FREEALL,7,"FREEALL",docode,0,0
	push	bc		; save TOS
	push	ix		; save RP
	ld	hl,pool_table
	ld	b,224
.freeloop
	ld	a,(hl)
	ld	(hl),0
	inc	hl
	and	a
	call	nz,freepool
	djnz	freeloop
	pop	ix		; restore RP
	pop	bc		; restore TOS
	next
.freepool
	ld	c,0
	add	a,a
	rl	c
	add	a,a
	rl	c
	add	a,a
	rl	c
	add	a,a
	rl	c
	ld	ixh,c		; ld ixh,c
	ld	ixl,a		; ld ixl,a
	call_oz(os_mcl)
	ret

;Z >FAR		addr -- dl-addr		Convert local address to far
;    0 ;
     head TOFAR,4,">FAR",docode,0,tofarn-tofarf
.tofarf push bc
	ld bc,0
.tofarn
	next

;Z C@L  	dl-addr -- char		Fetch character
     head CFETCHL,3,"C@L",docode,0,cfetln-cfetlf
.cfetlf ex      de,hl
	ex	(sp),hl		; save IP
	ex	de,hl		; CDE=far address
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
	call	faradd
	ld	c,(hl)
	ld	b,0		; TOS=char
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; restore seg 1
	pop	de		; restore IP
.cfetln
	next

;Z C!L		char L-addr --		Store character
     head CSTOREL,3,"C!L",docode,0,cstoln-cstolf
.cstolf ld      a,c
	pop	hl		; AHL=far address
	pop	bc		; C=character
	push	de		; save IP
	push	bc		; save character
	ld	c,a
	ex	de,hl		; CDE=far address
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
	call	faradd
	pop	bc
	ld	(hl),c		; store character
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; restore seg 1
	pop	de		; restore IP
	pop	bc		; new TOS
.cstoln
	next

;Z @L		L-addr -- x		Fetch cell
     head FETCHL,2,"@L",docode,0,fetcln-fetclf
.fetclf ex      de,hl
	ex	(sp),hl		; save IP
	ex	de,hl		; CDE=far address
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
	call	faradd
	ld	a,(hl)
	call	incfar
	ld	b,(hl)
	ld	c,a		; TOS=cell value
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; restore seg 1
	pop	de		; restore IP
.fetcln
	next

;Z !L		x L-addr --		store cell
     head STOREL,2,"!L",docode,0,storln-storlf
.storlf pop     hl              ; CHL=far address
	exx
	pop	bc		; BC'=word
	exx
	push	de		; save IP
	ex	de,hl		; CDE=far address
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
	call	faradd
	exx
	ld	a,c
	exx
	ld	(hl),a		; store low byte
	call	incfar
	exx
	ld	a,b
	exx
	ld	(hl),a		; store high byte
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; restore seg 1
	pop	de		; restore IP
	pop	bc		; new TOS
.storln
	next

;Z CMOVEL	dl-addr1 dl-addr2 u --	move data (far addresses)
     head CMOVEL,6,"CMOVEL",docode,0,cmvln-cmvlf
.cmvlf  push    bc
	pop	hl		; HL=count
	exx
	pop	bc
	pop	de		; C'D'E'=dl-addr2
	exx
	pop	bc
	ex	de,hl
        ex      (sp),hl         ; save IP
	ex	de,hl		; CDE=dl-addr1, HL=count
	push	hl
	ex	(sp),iy		; IY=count, save UP
	ld	a,(seg1)
	ex	af,af'		; save seg 1 binding
.cmovelloop
	ld	a,iyl		; ld a,iyl
	or	iyh		; or iyh
	jr	z,cmovelend	; on if finished
	call	faradd
	ld	a,(hl)		; char from dl-addr1
	push	af
	call	incfar
	exx
	call	faradd
	pop	af
	ld	(hl),a		; store at dl-addr2
	call	incfar
	exx
	dec	iy
	jr	cmovelloop
.cmovelend
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; restore seg 1
	pop	iy		; restore UP
	pop	de		; restore IP
	pop	bc		; new TOS
.cmvln
	next


; Package-handling words =================================

; The DEFERred word (RC_PNF) works as:  (RC_PNF) ( callid-- )
; It may display an error message asking access to be restored to the
; package, or THROW an exception (as the default action, PKGMIA does)


;Z PKGS?        -- flag         Check if package-handling is available
     head PKGSQ,5,"PKGS?",docode,0,pkgsqn-pkgsqf
.pkgsqf push    bc              ; stack previous TOS
	call_pkg(pkg_ayt)
	ld	a,0
	ccf
	sbc	a,a		; A=-1 if available, 0 if not
	ld	b,a
	ld	c,a		; new TOS=flag
.pkgsqn
	next

;Z AYT?		pkgid -- ver	Check if package is available (ver=0 if not)
     head AYTQ,4,"AYT?",docode,0,noayt-aytqf
.aytqf  push    bc
	ex	(sp),iy		; IYl=pkgid
	ld	iyh,$02		; ld iyh,$02 (xx_ayt)
	rst	$10		; make the call
	ld	bc,0		; ver=0 if not found
	call	nc,getpver	; only try to get version if found
	pop	iy
.noayt
	next

;Z PKGINF	pkgid -- 0-addr Dmem hdls res	Get package information
;  NOTE: Results of this call are not valid if package not installed,
;        so check with AYT? immediately before
     head PKGINF,6,"PKGINF",docode,0,0
	push	de		; save IP
	push	bc
	ex	(sp),iy		; IYl=pkgid
	ld	iyh,$00		; ld iyh,$00 (xx_inf)
	rst	$10
	jr	c,nopkginf
	ld	a,(seg1)
	ex	af,af'		; save seg1 binding
	ld	a,b
	ld	(seg1),a
	out	(segp1),a	; bind in bank containing name
	res	7,h
	set	6,h		; HL addressed in seg1
	ld	de,usr_zpad
.cpname
	ld	a,(hl)
	inc	hl
	ld	(de),a
	inc	de
	and	a
	jr	nz,cpname	; until null copied
	ex	af,af'
	ld	(seg1),a
	out	(segp1),a	; rebind seg1
	ld	iyh,$06		; ld iyh,$06 (xx_dat)
	rst	$10
.nopkginf
	pop	iy		; restore UP
	ld	hl,usr_zpad
	ex	(sp),hl		; HL=IP; stacked address of name
	push	de
	ld	e,b
	ld	b,0
	push	bc		; stacked Dmem
	ld	c,e
	push	bc		; stacked hdls
	ld	c,a		; TOS=res
	jp	nexthl_code

; Subroutine to get version number of package
; IN:	IYl=package ID
; OUT:	BC=version number

.getpver
	push	de		; save DE
	ld	iyh,$00		; ld iyh,$00 (xx_inf)
	rst	$10		; make the call
	ld	b,d
	ld	c,e		; BC=version number
	pop	de		; restore DE
	ret

;Z (RC_PNF)	--	RC_PNF action
;   DEFER (RC_PNF)
    head BRRCPNF,8,"(RC_PNF)",dodefer,0,0
	defw usr_pkgmia

;Z (PKGERR)	--	Package error action
;   DEFER (PKGERR)
    head BRPKGERR,8,"(PKGERR)",dodefer,0,0
	defw usr_pkgerr

;Z PKGERR	rc --	Default action for (PKGERR)
;    256 + NEGATE THROW ;
     head PKGERR,6,"PKGERR",docolon,0,0
	defw LIT,256,PLUS,NEGATE,THROW

;Z PKGMIA	callid --	Default action for (RC_PNF)
;    255 AND 512 + NEGATE THROW ;
     head PKGMIA,6,"PKGMIA",docolon,0,0
	defw LIT,255,AND,LIT,512,PLUS,NEGATE,THROW

;Z PKGCALL     ressize,#params,minver,reason,callid --
;				    define C-style package call
; IMPORTANT NOTE!!
;   This call assumes all input parameters to be single-cell; if any
;   are doubles (equivalent to C pointers or longs) their components
;   *must* be swapped on stack before entry to this routine, since
;   ANSI FORTH requires high cell to be above low cell on the stack;
;   however, Z88DK Small C does the reverse. Double-sized results are
;   automatically placed on the stack in the correct order for FORTH.

;   NSDS CREATE  , C, , 2* , C, DOES> (code fragment)
    head PKGCALL,7,"PKGCALL",docolon,0,0
        defw NSDS,CREATE,COMMA,CCOMMA,COMMA,TWOSTAR,COMMA,CCOMMA,XDOES

.dopkg	dec     ix
        ld	(ix+0),d	; save IP on return stack
        dec	ix
        ld	(ix+0),e
        pop	hl		; pointer to data field (call, #params*2)
	push	bc		; put TOS on physical stack
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl		; HL=address of data
        ld	e,(hl)
        inc	hl
        ld	d,(hl)		; DE=call ID
        inc	hl
        dec	ix
        ld	(ix+0),h	; save datapointer on return stack
        dec	ix
        ld	(ix+0),l
	dec	ix
	ld	a,iyh
	ld	(ix+0),a
	dec	ix
	ld	a,iyl
	ld	(ix+0),a	; save UP on return stack
	push	ix		; save RP (where C's retaddr would be)
.repkg  push    de
        pop	iy		; IY=package call ID
	ld	l,(ix+2)
	ld	h,(ix+3)
	ld	a,(hl)		; A=reason code (mostly for ZSOCK)
	rst	$10		; make the call
	jr	nc,pkgok
	cp	rc_pnf
	jr	nz,pkgbad	; if not rc_pnf, go to report error
	ld	iyh,$02		; ld iyh,$02
	rst	$10		; Are You There?!
	jr	c,badcall	; no, couldn't install
	call	getpver		; BC=package version
	push	bc
	ld	l,(ix+2)
	ld	h,(ix+3)
	inc	hl
	ld	c,(hl)
	inc	hl
	ld	b,(hl)		; BC=required version
	pop	hl
	and	a
	sbc	hl,bc
	jr	nc,repkg	; okay, if reinstalled at least min version
.badcall
	ld	a,(ix+0)
	ld	iyl,a
	ld	a,(ix+1)
	ld	iyh,a		; restore UP for running deferred word
	push	de		; save 2OS=call ID that failed
	ld	b,d
	ld	c,e		; also TOS=call ID
	ld	de,repkg2
	ld	hl,(usr_pkgmia)
	jp	(hl)		; execute DEFERrd word (RC_PNF)
.repkg2 defw      repkg2+2
	push	bc
	pop	de
	jr	repkg
.pkgbad pop     ix              ; restore RP
	push	de
	push	hl		; save results DEHL
	ld	c,a
	ld	b,0		; BC=error code
	ld	a,(ix+0)
	ld	iyl,a
	ld	a,(ix+1)
	ld	iyh,a		; restore UP for running deferred words
	ld	a,c		; get back error code
	scf
	call	vector		; handle RC_QUIT/RC_ESC/RC_DRAW as usual
	ld	de,pkgbd2
	ld	hl,(usr_pkgerr)
	jp	(hl)		; execute DEFERred word (PKGERR)
.pkgbd2 defw      pkgbd2+2
	push	bc		; new TOS to physical stack
	pop	hl
	pop	de		; restore results
	jr	pkgend
.pkgok  pop     ix              ; restore RP
.pkgend exx                     ; save call result in D'E'H'L'
	ld	a,(ix+0)
	ld	iyl,a
	inc	ix
	ld	a,(ix+0)
	ld	iyh,a		; restore UP
	inc	ix
	ld	l,(ix+0)
	inc	ix
	ld	h,(ix+0)
	inc	ix
	inc	hl		; skip reason code
	inc	hl
	inc	hl		; skip required version
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	inc	hl
	ex	de,hl
	add	hl,sp
	ld	sp,hl		; adjust SP to discard input parameters
	ld	a,(de)		; A=size of expected result (bytes)
	and	a
	jr	nz,getres
	pop	bc		; get new TOS for void functions
	jr	gotres
.getres exx                     ; DEHL=results
	ld	b,0
	ld	c,l
	dec	a
	jr	z,gotres	; 1-byte result in C
	ld	b,h
	dec	a
	jr	z,gotres	; 2-byte result in BC
	push	bc
	ld	b,0
	ld	c,e
	dec	a
	jr	z,gotres	; 3-byte result in C,TOS
	ld	b,d		; 4-byte result in BC,TOS
.gotres ld      e,(ix+0)
	inc	ix
	ld	d,(ix+0)	; restore IP
	inc	ix
	next

;Z PID		-- pid		; get current process ID
     head PID,3,"PID",docode,0,pidn-pidf
.pidf   
	call_pkg(pkg_pid)
	push	bc
	ld	c,a
	ld	b,0
.pidn
	next


